# coding=utf-8
"""
Monitors WAN connection
"""
import asyncio

import requests
import requests.exceptions

from esst import commands, core

LOGGER = core.MAIN_LOGGER.getChild(__name__)


async def wan_available(retry: int = 0):
    """

    Returns: True if connected to WAN

    """
    try:
        response = requests.get('http://google.com', timeout=2)
        commands.DCS.unblock_start('no WAN connection available')
        commands.DISCORD.can_start()
        return bool(response.ok)
    except requests.exceptions.RequestException:
        if retry < 5:
            LOGGER.debug(f'Internet connection loss detected, retry {retry}')
            await asyncio.sleep(2)
            result = await wan_available(retry + 1)
            return result
        LOGGER.debug(f'Internet connection loss detected, no more retry')
        commands.DISCORD.cannot_start()
        commands.DCS.block_start('no WAN connection available')
        return False


async def monitor_connection():
    """
    Loop that checks WAN every 5 seconds
    """
    LOGGER.debug('starting connection monitoring loop')

    while not core.CTX.exit:

        current_status = await wan_available()

        if current_status != core.CTX.wan:
            if current_status:
                LOGGER.debug('connected to the Internet')
                commands.DISCORD.say(
                    'I just lost internet connection, server is scheduled to be restarted')
            else:
                LOGGER.warning('internet connection lost !')
                commands.DCS.kill(force=False, queue=True)
            core.CTX.wan = current_status

        await asyncio.sleep(10)

    LOGGER.debug('end of connection monitoring loop')



# coding=utf-8
"""
Clean folders of old files
"""
import datetime
import os

import parsedatetime

from esst.core import CFG, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


def parse_age_string(age_str):
    """
    Converts age string to datetime timestamp

    Args:
        age_str: human friendly age string

    Returns: datetime timestamp

    """
    time_struct, parse_status = parsedatetime.Calendar().parse(age_str)
    if parse_status != 1:
        LOGGER.error(f'unable to parse age: {age_str}')
        return False
    return datetime.datetime(*time_struct[:6]).timestamp()


def remove_file_if_older_than(file_path, age):
    """
    Removes file if it's older than age

    Args:
        file_path: path to file to remove
        age: maximum age of file
    """
    file = os.path.abspath(file_path)
    if not os.path.exists(file):
        LOGGER.error(f'file does not exist: {file}')
        return
    modification_time = os.path.getmtime(file)
    # LOGGER.debug(f'"{file}" modification time: {modification_time}')
    if modification_time <= age:
        LOGGER.info(f'removing: {file}')
        os.unlink(file)


def _remove_old_files_from_folder(folder, age):
    LOGGER.info(f'cleaning folder "{folder}" of all files older than {age}')

    age = parse_age_string(age)
    if not age:
        return

    for root, _, files in os.walk(folder):
        for file in files:
            remove_file_if_older_than(os.path.join(root, file), age)


def clean_all_folder():
    """
    Cleans all folders according to current config
    """
    paths_to_clean = CFG.remove_files
    if paths_to_clean:  # pylint: disable=using-constant-test
        for remove_config in paths_to_clean:  # pylint: disable=not-an-iterable
            name = tuple(remove_config.keys())[0]
            LOGGER.info(f'processing: {name}')
            remove_config = remove_config[name]
            if 'folder' not in remove_config.keys():
                LOGGER.error(f'missing "folder" in {name}')
                return
            if 'age' not in remove_config.keys():
                LOGGER.error(f'missing "age" in {name}')
                return
            if not os.path.exists(remove_config['folder']):
                LOGGER.error(f'path does not exist: {remove_config["folder"]}')
                return
            _remove_old_files_from_folder(**remove_config)
    else:
        LOGGER.debug('no folder to clean')



# coding=utf-8
"""
Manages UR installation path
"""
import typing
from pathlib import Path

from esst import core, utils

from ._ur_status import Status as URStatus

try:
    import winreg
except ImportError:
    from unittest.mock import MagicMock

    winreg = MagicMock()

A_REG = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _get_ur_install_path_from_registry() -> typing.Union[Path, None]:
    LOGGER.debug('searching for base "Saved Games" folder')
    try:
        with winreg.OpenKey(A_REG, r"Software\sSoft\UniversRadio") as key:
            # noinspection SpellCheckingInspection
            return Path(winreg.QueryValueEx(key, "Install_Dir")[0])
    except FileNotFoundError:
        return None


def discover_ur_install_path():
    """
    Tries to find Saved Games on this system

    Returns: Saved Games dir
    """
    LOGGER.debug('discovering UR install path')
    if not core.CFG.ur_path:
        LOGGER.debug('no UR install path in Config, looking it up')
        _ur_install_path = _get_ur_install_path_from_registry()

    else:
        LOGGER.debug('UR install path found in Config')
        _ur_install_path = Path(core.CFG.saved_games_dir)
        if not _ur_install_path.is_dir():
            LOGGER.error(f'UR install path provided in config file is invalid: {_ur_install_path}')
            _ur_install_path = _get_ur_install_path_from_registry()

    LOGGER.debug(f'using Saved Games path: {_ur_install_path}')
    core.FS.ur_install_path = \
        URStatus.install_path = \
        _ur_install_path
    core.FS.ur_settings_folder = \
        URStatus.settings_folder = \
        Path(core.FS.saved_games_path, 'UniversRadio')
    core.FS.ur_voice_settings_file = \
        URStatus.voice_settings_file = \
        Path(core.FS.ur_settings_folder, 'VoiceService.dat')
    utils.create_simple_backup(core.FS.ur_voice_settings_file, file_must_exist=False)



# coding=utf-8
"""
Manages base classes for Sentry contexts
"""
import inspect


class SentryClassContext:
    """
    Base class for data class (class methods only)
    """

    @classmethod
    def get_context(cls) -> dict:
        """

        Returns: dict context for Sentry

        """
        return {
            member: value
            for member, value in inspect.getmembers(cls, lambda a: not inspect.ismethod(a))
            if not member.startswith('_')
        }


class SentryContext:
    """
    Base class for full blown class
    """

    def get_context(self) -> dict:
        """

        Returns: dict context for Sentry

        """
        return {
            member: value
            for member, value in inspect.getmembers(self, lambda a: not inspect.ismethod(a))
            if not member.startswith('_')
        }



# coding=utf-8
"""
Runs a Discord bot using the discord.py library
"""

import asyncio
import os
import random

import aiohttp
import aiohttp.errors
import discord
import websockets.exceptions

from esst.core import CFG, CTX, MAIN_LOGGER

from .abstract import AbstractDiscordBot, AbstractDiscordCommandParser
from .catch_exc import catch_exc
from .chat_commands.parser import make_root_parser
from .events import DiscordEvents
from .logging_handler import register_logging_handler
from .tasks import DiscordTasks

LOGGER = MAIN_LOGGER.getChild(__name__)


class App(DiscordTasks,  # pylint: disable=too-many-instance-attributes
          DiscordEvents,
          AbstractDiscordBot):
    """
    ESST Discord bot.

    This class is self-contained in a thread that will auto-start.
    """

    @property
    def parser(self) -> AbstractDiscordCommandParser:
        """Chat command parser"""
        return self._parser

    @property
    def channel(self) -> discord.Channel:
        """Active channel"""
        return self._channel

    @property
    def user(self) -> discord.User:
        """User object owned by the bot"""
        return self._user

    @property
    def member(self) -> discord.Member:
        """Member object owned bu the bot"""
        return self._member

    @property
    def server(self) -> discord.Server:
        """Server object for the bot"""
        return self._server

    @property
    def client(self) -> discord.Client:
        """Client object for the bot"""
        return self._client

    @property
    def ready(self) -> bool:
        """True when the bot is initialized"""
        return bool(self._ready)

    def __init__(self):
        self._parser = make_root_parser()
        self._client = None
        self._server = None
        self._user = None
        self._member = None
        self._channel = None
        self._ready = False
        self.tasks = None

        if not CTX.start_discord_loop:
            LOGGER.debug('skipping Discord bot startup')
            return

        LOGGER.debug('starting Discord bot')
        CTX.discord_msg_queue.put(CFG.discord_motd)
        register_logging_handler()

    def _create_client(self):
        self._client = discord.Client(loop=CTX.loop)
        self.client.on_ready = self.on_ready
        self.client.on_message = self.on_message
        self.client.on_message_edit = self.on_message_edit

    async def get_channel(self):
        """
        Sets the channel for the bot.

        If the channel does not exist, it will be created on the server, provided the bot has the authorization.
        """
        for channel in self.server.channels:
            if channel.name == CFG.discord_channel:
                self._channel = channel
                break
        else:
            # noinspection PyUnresolvedReferences
            self._channel = await self.client.create_channel(
                server=self.server,
                name=CFG.discord_channel,
                type=discord.ChannelType.text,
            )

    async def _update_profile(self, user_name: str = None):
        user_name = user_name or CFG.discord_bot_name
        profile_update = {}
        if self.user.name != user_name:
            profile_update['username'] = user_name
        if os.path.exists('avatar.png'):
            with open('avatar.png', 'rb') as handle:
                profile_update['avatar'] = handle.read()
        if profile_update:
            try:
                await self.client.edit_profile(**profile_update)
            except discord.errors.HTTPException:
                pass
            except aiohttp.errors.ClientResponseError:
                pass
            except websockets.exceptions.ConnectionClosed:
                pass
            except RuntimeError:
                pass

    @catch_exc
    async def _update_presence(
            self,
            status: str = discord.Status.online,
            afk: bool = False,
    ):
        members = list(self.server.members)
        members.remove(self._member)
        random_member = random.choice(members)
        await self.client.change_presence(
            game=discord.Game(
                name=f'with {random_member.display_name}',
                url=r'https://goo.gl/ZrxoaV',
                type=0,
            ),
            status=status,
            afk=afk,
        )

    @catch_exc
    async def on_ready(self):
        """
        Triggers when the bot is ready.
        """
        if not self.ready:
            self._user = self.client.user
            await self._update_profile()
            LOGGER.debug(f'Logged in as: {self.client.user.name}')
            try:
                self._server = set(self.client.servers).pop()
            except KeyError:
                LOGGER.error('Your discord bot has not server to connect to\n'
                             'Go to https://discordapp.com/developers/applications/me to create a bot, and note '
                             'the client ID.\n'
                             'Use the client ID in the following URL to join you bot to your Discord server:\n'
                             'https://discordapp.com/oauth2/authorize?client_id=CLIENT_ID&scope=bot')
            else:
                self._member = self.server.get_member(self.user.id)
                if self.user.display_name != CFG.discord_bot_name:
                    await self.client.change_nickname(self.member, CFG.discord_bot_name)
                await self._update_presence()

            await self.get_channel()

            self._ready = True

    @catch_exc
    async def _watch_for_exit_signals(self):
        if CTX.exit:
            while not self.ready:
                await asyncio.sleep(0.1)
            if self.ready:
                while not CTX.discord_msg_queue.empty():
                    await self._process_message_queue()
                await self.say('Bye bye !')
                await self.client.change_presence(status='offline')
            LOGGER.debug('closing Discord client')
            if self.client:
                while not self.client.is_logged_in:
                    await asyncio.sleep(0.1)
                if self.client.is_logged_in:
                    await self.client.logout()
                    await self.client.close()
                while not self.client.is_closed:
                    await asyncio.sleep(0.1)
            LOGGER.debug('Discord client is closed')
            return True

    async def watch_for_exit_signals(self):
        """
        Continuously runs and intercepts CTX.exit
        """
        while True:
            if await self._watch_for_exit_signals():
                break
            await asyncio.sleep(1)

    @catch_exc
    async def _run(self):
        if CTX.discord_can_start:
            LOGGER.debug('starting Discord client')
            self._create_client()
            await self.client.start(CFG.discord_token)
        else:
            await asyncio.sleep(1)

    @catch_exc
    async def run(self):
        """
        Main loop
        """

        if not CFG.discord_token:
            LOGGER.error('missing Discord token in config, cannot start bot')
            return

        if not CTX.start_discord_loop:
            LOGGER.debug('skipping Discord loop')
            return

        CTX.loop.create_task(self.watch_for_exit_signals())
        CTX.loop.create_task(self.monitor_queues())

        while not CTX.exit:
            await self._run()

        LOGGER.debug('end of Discord loop')



# coding=utf-8
"""
ATIS status
"""
import esst.utils.sentry_context


class Status(esst.utils.sentry_context.SentryClassContext):
    """
    ATIS status
    """

    def __init__(self):
        pass

    univers_radio = 'unknown'
    active_atis = 'unknown'



# coding=utf-8
"""
Decorator for Discord commands arguments
"""
from argh.constants import ATTR_ARGS

from esst import core

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def arg(*args, **kwargs):
    """
    Decorator for Discord commands arguments
    """

    def _wrapper(func):
        if 'protected' in kwargs:
            setattr(func, 'protected_', True)
            return func
        declared_args = getattr(func, ATTR_ARGS, [])
        # The innermost decorator is called first but appears last in the code.
        # We need to preserve the expected order of positional arguments, so
        # the outermost decorator inserts its value before the innermost's:
        declared_args.insert(0, dict(option_strings=args, **kwargs))
        setattr(func, ATTR_ARGS, declared_args)
        return func

    return _wrapper



# coding=utf-8
"""
Manages config params for dcs application
"""

from elib.config import ConfigProp

NAMESPACE = 'DCS'


class DCSConfig:
    """
    Manages config params for dcs application
    """

    dcs_path = ConfigProp(str, default='', namespace=NAMESPACE)
    dcs_idle_cpu_usage = ConfigProp(int, default='5', namespace=NAMESPACE)
    dcs_high_cpu_usage = ConfigProp(int, default='0', namespace=NAMESPACE)
    dcs_high_cpu_usage_interval = ConfigProp(int, default='5', namespace=NAMESPACE)
    dcs_ping_interval = ConfigProp(int, default='30', namespace=NAMESPACE)
    dcs_grace_period = ConfigProp(int, default='30', namespace=NAMESPACE)
    dcs_cpu_affinity = ConfigProp(list, default='', namespace=NAMESPACE)
    dcs_cpu_priority = ConfigProp(str, default='', namespace=NAMESPACE)
    dcs_delete_logs_older_than = ConfigProp(str, default='', namespace=NAMESPACE)
    dcs_can_start = ConfigProp(bool, default='true')



# coding=utf-8
"""
Manages a UDP socket and does two things:

1. Retrieve incoming messages from DCS and update :py:class:`esst.core.status.status`
2. Sends command to the DCS application via the socket
"""

import asyncio
import json
import socket
import time

from esst.core import CFG, CTX, MAIN_LOGGER, Status
from esst.utils import now

LOGGER = MAIN_LOGGER.getChild(__name__)

KNOWN_COMMANDS = ['exit dcs']

PING_TIMEOUT = 30


class DCSListener:
    """
    This class is a self-starting thread that creates and manages a UDP socket as a two-way communication with DCS
    """

    def __init__(self):
        self.monitoring = False
        self.last_ping = None
        self.startup_age = None

        if not CTX.start_listener_loop:
            LOGGER.debug('skipping startup of socket')
            return

        LOGGER.debug('starting socket thread')

        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.server_address = ('localhost', 10333)

        self.cmd_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.cmd_address = ('localhost', 10334)

    def _parse_ping(self, data: dict):
        if Status.paused != data.get('paused'):
            if not data.get('paused'):
                LOGGER.info('DCS server is ready!')
        players = data.get('players')
        if players != Status.players:
            players, old_players = set(players), set(Status.players)
            joined = players - old_players
            left = old_players - players
            if joined:
                LOGGER.info(f'player(s) joined: {", ".join(joined)}')
            if left:
                LOGGER.info(f'player(s) left: {", ".join(left)}')
        self.last_ping = time.time()
        Status.server_age = data.get('time')
        Status.mission_time = data.get('model_time')
        Status.paused = data.get('paused')
        Status.mission_file = data.get('mission_filename')
        Status.mission_name = data.get('mission_name')
        Status.players = data.get('players')

        CTX.players_history.append((now(), len(Status.players)))

    def _parse_status(self, data: dict):
        LOGGER.debug(f'DCS server says: {data["message"]}')
        if data['message'] in ['loaded mission']:
            LOGGER.debug('starting monitoring server pings')
            self.last_ping = time.time()
            self.monitoring = True
            CTX.listener_monitor_server_startup = False
        if data['message'] in ['stopping simulation']:
            LOGGER.debug('stopped monitoring server pings')
            self.monitoring = False
        Status.server_status = data['message']

    async def _parse_commands(self):
        await asyncio.sleep(0.1)
        if not CTX.listener_cmd_queue.empty():
            command = CTX.listener_cmd_queue.get_nowait()
            if command not in KNOWN_COMMANDS:
                raise ValueError(f'unknown command: {command}')
            else:
                command = {'cmd': command}
                command = json.dumps(command) + '\n'
                LOGGER.debug(f'sending command via socket: {command}')
                self.cmd_sock.sendto(command.encode(), self.cmd_address)

    async def _monitor_server(self):
        await asyncio.sleep(0.1)
        if self.monitoring:
            if time.time() - self.last_ping > CFG.dcs_ping_interval:
                LOGGER.error('It has been 30 seconds since I heard from DCS. '
                             'It is likely that the server has crashed.')
                CTX.dcs_do_restart = True
                self.monitoring = False

    async def _monitor_server_startup(self):
        await asyncio.sleep(0.1)
        if CTX.listener_monitor_server_startup:
            if self.startup_age is None:
                self.startup_age = time.time()
            if time.time() - self.startup_age > CFG.dcs_server_startup_time:
                LOGGER.error('DCS is taking more than 2 minutes to start a multiplayer server.\n'
                             'Something is wrong ...')
                CTX.listener_monitor_server_startup = False

    async def _read_socket(self):
        await asyncio.sleep(0.1)
        try:
            data, _ = self.sock.recvfrom(4096)
            data = json.loads(data.decode().strip())
            if data['type'] == 'ping':
                self._parse_ping(data)
            if data['type'] == 'status':
                self._parse_status(data)
            else:
                pass
        except socket.timeout:
            pass

    async def run(self):
        """
        Infinite loop that manages a UDP socket and does two things:

        1. Retrieve incoming messages from DCS and update :py:class:`esst.core.status.status`
        2. Sends command to the DCS application via the socket
        """
        if not CTX.start_listener_loop:
            LOGGER.debug('skipping startup of socket loop')
            return

        try:
            self.sock.bind(self.server_address)
        except socket.error as exc:
            if exc.errno == 10048:
                MAIN_LOGGER.error(
                    'cannot bind socket, maybe another instance of ESST is already running?')
                exit(-1)

        self.sock.settimeout(1)

        while not CTX.exit:
            await self._read_socket()
            await self._parse_commands()
            await self._monitor_server_startup()
            await self._monitor_server()
            await asyncio.sleep(0.1)

        self.sock.close()
        LOGGER.debug('end of listener loop')

    async def run_until_dcs_is_closed(self):
        """
        Parse commands until DCS has closed itself (or was killed)
        """
        while Status.dcs_application != 'not running':
            await self._parse_commands()
            await asyncio.sleep(0.1)
        self.cmd_sock.close()



# coding=utf-8
"""
Interface to the ATIS package
"""

import typing

from esst import core as _core
from ._atis_generator import generate_atis as _generate_atis
from ._atis_get_info import get_info_for_icao as _get_info_for_icao
from ._atis_status import Status as _ATISStatus
from ._univers_radio import Status as _URStatus

_LOGGER = _core.MAIN_LOGGER.getChild(__name__)


def generate_atis(
        metar_str: str,
        include_icao: typing.List[str] = None,
        exclude_icao: typing.List[str] = None):
    """
    Create ATIS from METAR

    Args:
        metar_str: METAR
        include_icao: list of ICAO to include
        exclude_icao: list of ICAO to exclude

    """
    _generate_atis(metar_str, include_icao, exclude_icao)


def get_info_from_icao(icao: str):
    """
    Retrieves ATIS information for ICAO

    Args:
        icao: ICAO string
    """
    _get_info_for_icao(icao)


def init_module():
    """
    Initialize the ATIS module
    """
    from ._univers_radio import discover_ur_install_path
    discover_ur_install_path()
    if _core.CTX.sentry:
        _core.CTX.sentry.register_context(context_name='ATIS', context_provider=_ATISStatus)
        _core.CTX.sentry.register_context(context_name='UR', context_provider=_URStatus)



# coding=utf-8
"""
Manages the local repositories and mission files
"""
import typing
import uuid
from pathlib import Path

from esst.core import FS, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


def _get_mission_folder(*paths: typing.Union[str, Path]) -> Path:
    """
    Returns a directory on the filesystem, creating it if necessary, and ensuring that no file exists at the location

    Args:
        *paths: list of paths to concatenate

    Returns: Path object

    """
    path = Path(*paths)
    if path.exists() and path.is_file():
        raise RuntimeError(f'path already exists but is a file: "{path}"')
    if not path.exists():
        LOGGER.debug(f'creating directory: {str(path)}')
        path.mkdir(parents=True)
    return path


def get_base_missions_folder() -> Path:
    """
    Returns the folder in which all ESST related missions are contained

    Returns: Path object

    """
    return FS.dcs_mission_folder


def get_auto_missions_folder() -> Path:
    """
    Returns the folder in which missions modified by ESST are stored

    Returns: Path object

    """
    return _get_mission_folder(get_base_missions_folder(), 'AUTO')


def get_random_auto_mission_name(source_mission: Path) -> Path:
    """
    Creates a random name for a mission to be modified

    Args:
        source_mission: path to the source mission

    Returns: random mission path

    """
    if not source_mission.is_file():
        raise RuntimeError(f'source_mission if not a file: {source_mission}')
    if source_mission.suffix != '.miz':
        raise RuntimeError(
            f'source_mission if not a MIZ file: {source_mission}')
    _id = str(uuid.uuid4())[:8]
    _tmp_name = source_mission.stem
    _tmp_path = Path(get_auto_missions_folder(), f'{_tmp_name}_{_id}.miz')
    return _tmp_path


def list_missions():
    """
    List all missions available for loading
    """
    for file in get_auto_missions_folder().iterdir():
        yield file


def clean():
    """
    Removes all mission files contained in the auto folder
    """
    for file in get_auto_missions_folder().iterdir():
        yield file



# coding=utf-8
"""
Manages Discord chat commands
"""
import discord

from esst.core import CFG, MAIN_LOGGER
from esst.dcs import missions_manager
from esst.discord_bot import abstract

LOGGER = MAIN_LOGGER.getChild(__name__)


# noinspection PyAbstractClass
class DiscordEvents(abstract.AbstractDiscordBot):  # pylint: disable=abstract-method
    """
    Manages Discord chat commands
    """

    async def on_message_edit(self, _: discord.Message, after: discord.Message):
        """
        Dummy event catcher to re-transmit message when they're edited

        Messages are simply re-routed to the "on_message" event

        Args:
            _: message before edit
            after: message after edit
        """
        await self.on_message(after)

    # pylint: disable=too-many-branches
    async def on_message(self, message: discord.Message):  # noqa: C901
        """
        Triggers on any message received from the Discord server

        Args:
            message: message received

        """
        if message.author.id == self.member.id:
            return
        if message.channel != self.channel:
            return
        if CFG.discord_admin_role:
            is_admin = bool([role for role in CFG.discord_admin_role  # pylint: disable=not-an-iterable
                             if role in [role.name for role in message.author.roles]])
        else:
            is_admin = True
        if message.attachments:
            for attach in message.attachments:
                if attach['filename'].endswith('.miz'):
                    if not is_admin:
                        LOGGER.error(
                            f'only users with role "{CFG.discord_admin_role}" can load missions on the server')
                        return
                    overwrite = 'overwrite' in message.content
                    load = 'load' in message.content
                    force = 'force' in message.content
                    missions_manager.download_mission_from_discord(attach, overwrite, load, force)
        if message.content.startswith('!'):
            LOGGER.debug(f'received "{message.content}" command from: {message.author.display_name}'
                         f'{" (admin)" if is_admin else ""}')

            self.parser.parse_discord_message(message.content, is_admin)



# coding=utf-8
"""
Dummy abstract class to help split the Discord bot into multiple more manageable classes
"""
import abc
import sys

import discord


class AbstractDiscordCommandParser:
    """
    Generic command parser for Discord bot
    """
    @abc.abstractmethod
    def dispatch(self,
                 argv=None,
                 add_help_command=True,
                 completion=True,
                 pre_call=None,
                 output_file=sys.stdout,
                 errors_file=sys.stderr,
                 raw_output=False,
                 namespace=None,
                 skip_unknown_args=False):
        """
        Dispatch a command to an underlying function

        Args:
            argv: list of arguments
            add_help_command: whether or not to add help command
            completion: auto-completion of missing commands
            pre_call: function to call before
            output_file: redirect stdout
            errors_file: redirect stderr
            raw_output: outputs result as text
            namespace: command namespace
            skip_unknown_args: ignore unknown args
        """
        pass

    @abc.abstractmethod
    def parse_args(self, args=None, namespace=None):
        """
        Parse received args

        Args:
            args: arg list
            namespace: command namespace
        """
        pass

    @abc.abstractmethod
    def format_help(self):
        """
        Format and return help text

        """
        pass

    @abc.abstractmethod
    def parse_discord_message(self, message: str, is_admin: bool):
        """
        PArses message from Discord

        Args:
            message: message content
            is_admin: is sender an admin?

        """
        pass


class AbstractDiscordBot:
    """
    Dummy abstract class to help split the Discord bot into multiple more manageable classes
    """

    @property
    @abc.abstractmethod
    def parser(self) -> AbstractDiscordCommandParser:
        """
        Command parser
        """
        pass

    @property
    @abc.abstractmethod
    def ready(self) -> bool:
        """
        Indicates if the bot is ready to process messages

        Returns: readiness as a boolean
        """
        pass

    @property
    @abc.abstractmethod
    def client(self) -> discord.Client:
        """
        Represents the bot :py:class:`discord.Client` object
        """
        pass

    @property
    @abc.abstractmethod
    def server(self) -> discord.Server:
        """
        Represents the bot :py:class:`discord.Server` object
        """
        pass

    @property
    @abc.abstractmethod
    def user(self) -> discord.User:
        """
        Represents the bot :py:class:`discord.User` object
        """
        pass

    @property
    @abc.abstractmethod
    def member(self) -> discord.Member:
        """
        Represents the bot :py:class:`discord.Member` object
        """
        pass

    @property
    @abc.abstractmethod
    def channel(self) -> discord.Channel:
        """
        Represents the bot :py:class:`discord.Channel` object

        This is the channel the bot will listen for commands in
        """
        pass

    @abc.abstractmethod
    def say(self, message: str):
        """
        Sends a message to the channel

        Args:
            message: message to send as string
        """
        pass

    @abc.abstractmethod
    def send(self, file_path: str):
        """
        Sends a file to the channel

        Args:
            file_path: path to file to send
        """
        pass



# coding=utf-8
"""
Manages GameGUI hooks
"""

import elib

from esst.core import CTX, FS, MAIN_LOGGER
from esst.utils import read_template

LOGGER = MAIN_LOGGER.getChild(__name__)


# noinspection SpellCheckingInspection


def _remove_old_file():
    old_file_path = elib.path.ensure_file(FS.variant_saved_games_path, 'DCS/Scripts/ESSTGameGUI.lua', must_exist=False)

    if old_file_path.exists():
        LOGGER.debug('removing old GameGUI script')
        old_file_path.unlink()


def _install_hook():
    hook_folder = elib.path.ensure_dir(FS.variant_saved_games_path, 'Scripts/Hooks', must_exist=False, create=True)
    LOGGER.debug(f'hooks folder: {hook_folder}')
    esst_hook_path = elib.path.ensure_file(hook_folder, 'esst.lua', must_exist=False)
    LOGGER.debug(f'ESST hook path: {esst_hook_path}')
    elib.path.ensure_dir(esst_hook_path.parent, must_exist=False, create=True)
    LOGGER.debug('writing ESST hook to file')
    with open(esst_hook_path, 'w') as stream:
        stream.write(read_template('game_gui.lua'))


def install_game_gui_hooks():
    """
    Installs the GameGUI hooks in DCS Scripts folder
    """
    _remove_old_file()

    if CTX.dcs_install_hooks:
        LOGGER.debug('installing GameGUI hooks')
        _install_hook()
    else:
        LOGGER.debug('skipping installation of GameGUI hooks')



# coding=utf-8
# pylint: disable=too-few-public-methods
"""
Contains all objects relative to UR ATIS settings
"""
import re

import inflect


class URLinkType:
    """
    Setting that represents the type of link to an MP3 file, either local (on the system's hard drive) or remote (URL)
    """
    local_file = '1'
    url = '2'


class URFrequency:
    """
    Frequency for the ATIS

    Format: XXX.XXX;MOD
    Where XXX.XXX is the frequency and MOD is one of:
        A: AM
        F: FM
        S: SATCOM

    """
    ur_freq_re = re.compile(r'^\d{3}\.\d{3};[ASF]$')

    def __init__(self, freq: str):
        if not URFrequency.ur_freq_re.match(freq):
            raise ValueError()
        self.freq = freq

    def __str__(self):
        return self.freq

    def long_freq(self):
        """Returns nice frequency (123.456 AM)"""
        freq, mod = self.freq.split(';')
        if mod == 'S':
            mod = 'SATCOM'
        else:
            mod = f'{mod}M'
        return f'{freq} {mod}'


class URCoord:
    """
    Coordinates for an airfield (includes elevation)
    """
    ur_coord_re = re.compile(r'^\d{2}\.\d{6}$')
    ur_elev_re = re.compile(r'^\d+$')

    def __init__(self, lat: str, long: str, elev: str):
        if not URCoord.ur_coord_re.match(lat):
            raise ValueError(f'invalid latitude: {lat}')
        if not URCoord.ur_coord_re.match(long):
            raise ValueError(f'invalid longitude: {long}')
        if not URCoord.ur_elev_re.match(elev):
            raise ValueError(f'invalid elevation: {elev}')
        self.lat = lat
        self.long = long
        self.elev = str(int(elev) + 500)

    def __str__(self):
        return f'{self.lat};{self.long};{self.elev}'


class Runway:

    """
    Represents one or multiple runways
    """

    def __init__(self, heading: str):
        if 'L' in heading.upper():
            self._qualifier = 'L'
            self._long_qualifier = ' left'
            self._heading = heading.replace('L', '')
        elif 'R' in heading.upper():
            self._qualifier = 'R'
            self._long_qualifier = ' right'
            self._heading = heading.replace('R', '')
        else:
            self._qualifier = ''
            self._long_qualifier = ''
            self._heading = heading
        self._long_heading = ' '.join([inflect.engine().number_to_words(number) for number in self._heading])

    def __int__(self):
        try:
            return int(f'{self._heading}0')
        except ValueError:
            raise ValueError(self._heading)

    def __str__(self):
        return f'{self._heading}{self._qualifier}'

    def long_name(self):
        """

        Returns: long name of the airfield for use in ATIS speech

        """
        return f'{self._long_heading}{self._long_qualifier}'


# pylint: disable=too-many-arguments,too-few-public-methods
class Airfield:
    """
    Represents an airfield, complete with ICAO, name, ATIS frequency and available runways
    """

    def __init__(self, icao: str, name: str, coord: URCoord, atis_freq: URFrequency, runways: str):
        self.atis_freq = atis_freq
        self.coord = coord
        self.icao = icao
        self.name = name
        self.runways = []
        for runway in runways.split(','):
            self.runways.append(Runway(runway))

    def get_active_runway(self, wind_direction: int) -> Runway:
        """
        Returns the active runway depending on wind

        Args:
            wind_direction: wind direction in degrees

        Returns: active runway as a Runway instance

        """

        def diff_angle(angle1, angle2):
            """
            Computes the difference in degrees between two angles

            Args:
                angle1: first angle
                angle2: second angle

            Returns: absolute difference in degrees between the two angles

            """
            return abs((angle1 - angle2 + 180) % 360 - 180)

        result = []
        for runway in self.runways:
            result.append((diff_angle(int(runway), wind_direction), runway))

        return sorted(result, key=lambda x: x[0])[0][1]



# coding=utf-8
"""
Manages DCS autoexec.cfg file
"""

from esst import core, utils

_LOGGER = core.MAIN_LOGGER.getChild(__name__)

_SILENT_CRASH_REPORT = '''\ncrash_report_mode = "silent"\n'''


def inject_silent_crash_report() -> bool:
    """
    Injects code needed for the new login method in MissionEditor.lua

    Args:
        dcs_path: path to the installation of DCS

    Returns:
        Bool indicating success of the operation

    """

    core.FS.ensure_path(core.FS.saved_games_path, 'saved games')
    autoexec_path = core.FS.ensure_path(core.FS.dcs_autoexec_file, 'dcs autoexec file', must_exist=False)
    utils.create_versioned_backup(autoexec_path, file_must_exist=False)

    if autoexec_path.exists():
        _LOGGER.debug('autoexec.cfg already exists, reading')
        content = autoexec_path.read_text(encoding='utf8')
    else:
        _LOGGER.debug('autoexec.cfg does not exist, creating')
        content = ''

    if _SILENT_CRASH_REPORT in content:
        _LOGGER.debug('silent crash report already enabled')
        return True

    content = f'{content}{_SILENT_CRASH_REPORT}'
    _LOGGER.debug(f'writing new "autoexec.cfg" content: {content}')
    autoexec_path.write_text(content)
    return True



# coding=utf-8
"""
Represents the global context
"""

import inspect
from collections import deque
from concurrent.futures import ProcessPoolExecutor
from queue import Queue


# TODO: Split context into the relevant modules
class Context:  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """
    Represents the global context
    """
    @classmethod
    def get_context(cls) -> dict:
        """

        Returns: dict context for Sentry

        """
        return {
            member: value
            for member, value in inspect.getmembers(cls, lambda a: not inspect.ismethod(a))
            if not (member.startswith('_') or 'history' in member)
        }

    exit = False
    restart = False
    loop = None
    sentry = None
    wan = False
    process_pool = ProcessPoolExecutor(max_workers=1)

    start_listener_loop = True
    start_discord_loop = True
    start_server_loop = True
    start_dcs_loop = True

    discord_can_start = False
    discord_msg_queue = Queue()
    discord_file_queue = Queue()

    dcs_install_hooks = True
    dcs_setup_dedi_config = True
    dcs_auto_mission = True

    dcs_can_start = False
    dcs_blocker = []
    dcs_show_cpu_usage = False
    dcs_show_cpu_usage_once = False
    dcs_do_kill = False
    dcs_do_queued_kill = False
    dcs_do_restart = False
    dcs_cpu_history = deque(maxlen=17280)
    dcs_mem_history = deque(maxlen=17280)

    listener_cmd_queue = Queue()
    listener_monitor_server_startup = False

    server_show_cpu_usage = False
    server_show_cpu_usage_once = False
    server_cpu_history = deque(maxlen=17280)
    server_mem_history = deque(maxlen=17280)
    server_bytes_sent_history = deque(maxlen=17280)
    server_bytes_recv_history = deque(maxlen=17280)

    players_history = deque(maxlen=17280)

    atis_speech = ''



# coding=utf-8
"""
Holds current state of the mission manager
"""


# class MissionManagerOperation:
#
#     def __init__(self, operation_name):
#         self.name = operation_name
#
#
# INIT = MissionManagerOperation('init')
# LOADING = MissionManagerOperation('loading')
# LOADED = MissionManagerOperation('loaded')
#
#
# class MissionManagerState:
#     """
#     Holds current state of the mission manager
#     """
#
#     current_operation = INIT
#     current_context = None



# coding=utf-8
"""
Manages UR settings
"""

from pathlib import Path

from esst import core

from ._ur_object import Airfield, URLinkType

LOGGER = core.MAIN_LOGGER.getChild(__name__)


class URVoiceServiceSettings:
    """
    Manages UR settings
    """

    def __init__(self):
        LOGGER.debug('creating UR settings instance')
        self._stations = []

    def add_station(self, airfield: Airfield):
        """
        Adds a station

        Args:
            airfield: station to add

        """
        LOGGER.debug(f'adding {airfield.icao} to UR settings')
        mp3_file = Path(f'./{airfield.icao}.mp3')
        station_line = ';'.join([
            str(URLinkType.local_file),
            str(airfield.atis_freq),
            str(airfield.coord),
            f'[{str(mp3_file.absolute())}]'
        ])
        LOGGER.debug(f'station line: {station_line}')
        self._stations.append(station_line)

    def write_settings_file(self):
        """
        Writes currently known station to UR settings file
        """
        LOGGER.debug(f'writing UR settings to: {core.FS.ur_voice_settings_file}')
        stations = '\n'.join(self._stations)
        full_text = f'Start of VSS DB\n{stations}\nEnd of VSS DB'
        core.FS.ur_voice_settings_file.write_text(full_text)



# coding=utf-8
"""
Connection with sentry.io
"""


import logging
import sys

import certifi
import raven
import raven.breadcrumbs
import raven.conf
import raven.handlers.logging

from esst import __version__
from esst.core import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


class Sentry(raven.Client):
    """
    Connection with sentry.io
    """

    def __init__(self, dsn):
        LOGGER.info('initializing Sentry')
        self.registered_contexts = {}
        raven.Client.__init__(
            self,
            f'{dsn}?ca_certs={certifi.where()}',
        )
        if self.is_enabled():
            LOGGER.info('Sentry is ready')

    def set_context(self):
        """
        Sets Sentry context

        """
        self.tags_context(
            dict(
                platform=sys.platform,
                version=__version__,
            )
        )
        try:
            self.tags_context(dict(windows_version=sys.getwindowsversion()))
        except AttributeError:
            pass

    def register_context(self, context_name: str, context_provider):
        """Registers a context to be read when a crash occurs; obj must implement get_context()"""
        LOGGER.debug(
            'registering context with Sentry: {}'.format(context_name))
        self.registered_contexts[context_name] = context_provider

    @staticmethod
    def add_crumb(message, category, level):
        """
        Manually add a crumb to the logger

        Args:
            message: message str
            category: category
            level: logging level

        """
        raven.breadcrumbs.record(
            message=message, category=category, level=level)

    def captureMessage(self, message, **kwargs):  # noqa: N802
        """
        Manually add a message

        Args:
            message: message content
            **kwargs: meh

        """
        self.set_context()
        if kwargs.get('data') is None:
            kwargs['data'] = {}
        if kwargs['data'].get('level') is None:
            kwargs['data']['level'] = logging.DEBUG
        for context_name, context_provider in self.registered_contexts.items():
            self.extra_context({context_name: context_provider.get_context()})
        super(Sentry, self).captureMessage(message, **kwargs)

    def captureException(self, exc_info=None, **kwargs):
        """Captures an exception"""
        self.set_context()

        LOGGER.debug('capturing exception')
        for k, context_provider in self.registered_contexts.items():
            self.extra_context({k: context_provider.get_context()})
        super(Sentry, self).captureException(exc_info, **kwargs)


# noinspection PyUnusedLocal
def filter_breadcrumbs(_logger, level, msg, *args, **kwargs):  # pylint: disable=unused-argument
    """
    Intercepts logging messages

    Args:
        _logger: originating logger
        level: record level
        msg: record message
        *args: logging args
        **kwargs: logging kwargs

    """
    skip_lvl = []
    skip_msg = []

    if level in skip_lvl or msg in skip_msg:
        return False

    if _logger == 'requests':
        return False
    return True


raven.breadcrumbs.register_special_log_handler('ESST', filter_breadcrumbs)



# coding=utf-8
"""
Manages missions for the server
"""

import pprint
import typing
from pathlib import Path

import emiz.weather
import humanize
import requests
from jinja2 import Template

from esst import atis, commands, core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _get_mission_folder() -> Path:
    return core.FS.dcs_mission_folder


def _get_auto_mission_folder() -> Path:
    auto_mission_folder = Path(_get_mission_folder().joinpath('AUTO'))
    if not auto_mission_folder.exists():
        LOGGER.debug(f'creating directory: {auto_mission_folder}')
        auto_mission_folder.mkdir(parents=True)
    return auto_mission_folder


class MissionPath:
    """
    Represents a MIZ file managed by ESST
    """

    def __init__(self, mission: typing.Union[str, Path]):
        self._path = Path(mission)
        self._orig_name = self._path.stem

    @property
    def name(self) -> str:
        """
        Returns: path's stem
        """
        return self._path.stem

    @property
    def orig_name(self):
        """
        Returns: original name of the mission
        """
        return self._orig_name

    @property
    def auto(self) -> 'MissionPath':
        """
        Returns: MissionPath object for the auto mission
        """
        if self._path.parent == _get_auto_mission_folder():
            return self

        return MissionPath(Path(_get_auto_mission_folder()).joinpath(self._path.name))

    @property
    def path(self) -> Path:
        """
        Returns: path to the MIZ file
        """
        return self._path

    def set_as_active(self, metar: str = None):
        """
        Write the settings file to set this mission as active
        Args:
            metar: metar string; if not provided,  will be inferred from MIZ file

        """

        LOGGER.info(f'setting active mission to: {self.name}')
        if not self:
            LOGGER.error(f'mission file not found: {self.path}')
            return
        template_option = dict(
            mission_file_path=str(self.path).replace('\\', '/'),
            passwd=core.CFG.dcs_server_password,
            name=core.CFG.dcs_server_name,
            max_players=core.CFG.dcs_server_max_players,
        )
        LOGGER.debug(f'rendering settings.lua template with options\n{pprint.pprint(template_option)}')
        content = Template(utils.read_template('settings.lua')).render(**template_option)
        settings_file = _get_settings_file_path()
        LOGGER.debug(f'settings file path: {settings_file}')
        utils.create_versioned_backup(settings_file)
        settings_file.write_text(content)

        if metar is None:
            LOGGER.debug(f'building metar from mission: {self.name}')
            # noinspection SpellCheckingInspection
            metar = emiz.weather.mizfile.get_metar_from_mission(str(self.path), icao='XXXX')
            LOGGER.info(f'metar for {self.name}:\n{metar}')
        atis.generate_atis(metar)
        core.Status.metar = metar

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return f'MissionPath({self._path})'

    def __bool__(self):
        return self.path.exists()


def _get_settings_file_path() -> Path:
    return core.FS.dcs_server_settings
    # return Path(core.FS.saved_games_path, 'DCS/Config/serverSettings.lua')


def set_active_mission(mission: str, metar: str = None):
    """
    Sets the mission as active in "serverSettings.lua"

    Args:
        mission: path or name of the MIZ file
        metar: METAR string for this mission
    """
    LOGGER.debug(f'setting active mission: {mission}')
    if metar:
        LOGGER.debug(f'using METAR: {metar}')
    mission = MissionPath(mission)
    mission.set_as_active(metar)


def delete(mission: MissionPath):
    """
    Removes a mission from the filesystem.

    Also removes leftover AUTO artifacts

    Args:
        mission: MissionPath instance to remove

    """
    if mission:
        LOGGER.info(f'removing: {mission.path}')
        mission.path.unlink()
    if mission.auto:
        LOGGER.info(f'removing: {mission.auto.path}')
        mission.auto.path.unlink()


def get_latest_mission_from_github():
    """
    Downloads the latest mission from a Github repository

    The repository needs to have releases (tagged)
    The function will download the first MIZ file found in the latest release
    """
    if core.CTX.dcs_auto_mission:
        LOGGER.debug('getting latest mission from Github')
        commands.DCS.block_start('loading mission')
        if core.CFG.auto_mission_github_repo and core.CFG.auto_mission_github_owner:
            LOGGER.debug('looking for newer mission file')
            latest_version, asset_name, download_url = utils.get_latest_release(
                core.CFG.auto_mission_github_owner, core.CFG.auto_mission_github_repo
            )
            LOGGER.debug(f'latest release: {latest_version}')
            local_file = MissionPath(Path(_get_mission_folder(), f'{asset_name}'))
            if not local_file:
                LOGGER.info(f'downloading new mission: {asset_name}')
                req = requests.get(download_url)
                if req.ok:
                    local_file.path.write_bytes(req.content)
                    local_file.set_as_active()
                else:
                    LOGGER.error('failed to download latest mission')
        else:
            LOGGER.error('no config values given for [auto mission]')
        commands.DCS.unblock_start('loading mission')
    else:
        LOGGER.debug('skipping mission update')


def download_mission_from_discord(discord_attachment,
                                  overwrite: bool = False,
                                  load: bool = False,
                                  force: bool = False):
    """
    Downloads a mission from a discord message attachment

    Args:
        force: force restart even with players connected
        discord_attachment: url to download the mission from
        overwrite: whether or not to overwrite an existing file
        load: whether or not to restart the server with the downloaded mission
    """
    url = discord_attachment['url']
    size = discord_attachment['size']
    filename = discord_attachment['filename']
    local_file = MissionPath(Path(_get_mission_folder(), filename))

    overwriting = ''
    if local_file:
        if overwrite:
            overwriting = ' (replacing existing file)'
        else:
            LOGGER.warning(f'this mission already exists: {local_file.path}\n'
                           f'use "overwrite" to replace it')
            return

    LOGGER.info(
        f'downloading: {filename} ({humanize.naturalsize(size)}) {overwriting}')
    with requests.get(url) as response:
        local_file.path.write_bytes(response.content)

    if load:
        if commands.DCS.there_are_connected_players() and not force:
            LOGGER.error('there are connected players; cannot restart the server now (use "force" to kill anyway)')
            return
        LOGGER.info(f'restarting the server with this mission')
        local_file.set_as_active()
        commands.DCS.restart(force=force)
    else:
        LOGGER.info(f'download successful, mission is now available')


def list_available_missions():
    """
    Generator that yields available mission in ESST's mission dir
    """
    count = 1
    for file in _get_mission_folder().glob('*.miz'):
        yield count, file
        count += 1


def get_running_mission() -> typing.Union['MissionPath', str]:
    """

    Returns: currently running mission as a MissionPath instance

    """
    mission = None
    if core.Status.mission_file and core.Status.mission_file != 'unknown':
        mission_path = Path(core.Status.mission_file)
        if mission_path.parent == 'AUTO':
            mission_path = Path(mission_path.parent.parent, mission_path.name)
        mission = MissionPath(mission_path)

    else:
        try:
            dcs_settings = _get_settings_file_path().read_text()
        except FileNotFoundError:
            LOGGER.error('please start a DCS server at least once before using ESST')
            exit(1)
        else:
            for line in dcs_settings.split('\n'):
                if '[1]' in line:
                    mission = MissionPath(line.split('"')[1])
                    break

    if mission:
        LOGGER.debug(f'returning active mission: {mission.name}')
        return mission

    LOGGER.error(f'current mission is "{mission.path}", but that file does not exist')
    return ''


def initial_setup():
    """
    Runs at the start of the DCS loop, to initialize the first mission
    """
    LOGGER.debug('initializing first mission')
    mission = get_running_mission()
    if isinstance(mission, MissionPath):
        LOGGER.info(f'building METAR for initial mission: {mission.orig_name}')
        metar = emiz.weather.mizfile.get_metar_from_mission(str(mission.path))
        core.Status.metar = metar
        atis.generate_atis(metar)
    else:
        LOGGER.error('no initial mission found')



# coding=utf-8
"""
Manages background tasks for Discord bot
"""

import asyncio
import time

from esst.core import CTX, MAIN_LOGGER

from .abstract import AbstractDiscordBot
from .catch_exc import catch_exc

LOGGER = MAIN_LOGGER.getChild(__name__)


# noinspection PyAbstractClass
class DiscordTasks(AbstractDiscordBot):  # pylint: disable=abstract-method
    """
    Abstract class that contains background tasks for :py:class:`esst.discord_bot.DiscordBot`
    """

    @catch_exc
    async def say(self, message: str):
        """
        Sends a message

        Args:
            message: message to send
        """
        time_stamp = time.strftime('%X')
        content = f'{time_stamp}```{message}```'
        if self.client and self.client.is_logged_in and self.channel:
            # noinspection PyUnresolvedReferences
            await self.client.send_message(self.channel, content=content)
            return True

    @catch_exc
    async def send(self, file_path: str):
        """Sends a file to a Discord channel"""
        if self.client and self.client.is_logged_in and self.channel:
            # noinspection PyUnresolvedReferences
            await self.client.send_file(self.channel, file_path, content='There you go:')
            return True

    async def _process_message_queue(self):
        if self.client.is_closed or not CTX.wan:
            return
        if not CTX.discord_msg_queue.empty():
            message = CTX.discord_msg_queue.get_nowait()
            if not await self.say(message):
                CTX.discord_msg_queue.put(message)

    async def _process_file_queue(self):
        if self.client.is_closed or not CTX.wan:
            return
        if not CTX.discord_file_queue.empty():
            file = CTX.discord_file_queue.get_nowait()
            if not await self.send(file):
                CTX.discord_msg_queue.put(file)

    @catch_exc
    async def _monitor_queues(self):
        while self.client is None:
            if CTX.exit:
                return True
            await asyncio.sleep(0.1)
        while not self.ready:
            if CTX.exit:
                return True
            await asyncio.sleep(0.1)
        await self.client.wait_until_ready()
        while not self.client.is_closed:
            await self._process_message_queue()
            await self._process_file_queue()
            await asyncio.sleep(0.1)
            if CTX.exit:
                return True

    async def monitor_queues(self):
        """
        Checks the message queue for pending messages to send
        """
        while not CTX.exit:
            if await self._monitor_queues():
                break



# coding=utf-8
"""
Validates config object
"""

import logging

from ._new_config import ESSTConfig

MAIN_LOGGER = logging.getLogger('ESST')

MANDATORY_CONFIG_OPTIONS = [
    'discord_token',
    'discord_channel',
    'discord_bot_name',
    'dcs_server_password',
    'dcs_server_name',
    'dcs_path',
]


def validate_config(config_instance: ESSTConfig):
    """
    Validates config object
    """
    MAIN_LOGGER.debug('validating config')
    missing_values = []
    for attrib_name in dir(config_instance):
        if not attrib_name.startswith('_'):
            attrib = getattr(config_instance, attrib_name)
            if not callable(attrib):
                if attrib_name in MANDATORY_CONFIG_OPTIONS and not attrib:
                    missing_values.append(attrib_name)

    if missing_values:
        missing_values = '\t' + '\n\t'.join(missing_values)
        MAIN_LOGGER.error(
            f"""
The following values are missing from your configuration:

{missing_values}

There is a "esst.yml.example" file provided with this repository to serve as a base to create your own config.
Refer to https://github.com/132nd-vWing/ESST/blob/master/README.md for more information about setting ESST up.""")
        exit(1)



# coding=utf-8
"""
Commands for the socket listener
"""

from esst.core import CTX, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


class LISTENER:
    """
    Commands for the socket listener
    """
    @staticmethod
    def monitor_server_startup_start():
        """Starts monitoring DCS startup"""
        LOGGER.debug('monitor server startup: start')
        CTX.listener_monitor_server_startup = True

    @staticmethod
    def monitor_server_startup_stop():
        """Stops monitoring server startup"""
        LOGGER.debug('monitor server startup: stop')
        CTX.listener_monitor_server_startup = False

    @staticmethod
    def exit_dcs():
        """Sends a socket signal to stop DCS"""
        LOGGER.debug('sending socket message to close DCS')
        CTX.listener_cmd_queue.put('exit dcs')



# coding=utf-8
"""
Manages Github requests
"""

import json
import typing

import requests

from esst.core import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)

BASE_URL = 'https://api.github.com/'


def _make_request(endpoint):
    url = f'{BASE_URL}{endpoint}'
    LOGGER.debug(url)
    req = requests.get(url)
    if req.ok:
        return json.loads(req.text or req.content)
    if req.text or req.content:
        resp = json.loads(req.text or req.content)
        if resp['message']:
            raise ConnectionError(
                f'Request failed: {url}\nMessage: {resp["message"]}')

    raise ConnectionError(f'Request failed: {url}')


def get_latest_release(owner: str, repo: str) -> typing.Tuple[str, str, str]:
    """

    Args:
        owner: owner of the Github repo
        repo: name of the Github repo

    Returns: latest version, asset name, asset download URL

    """
    resp = _make_request(f'repos/{owner}/{repo}/releases/latest')
    return resp['tag_name'], resp['assets'][0]['name'], resp['assets'][0]['browser_download_url']



# coding=utf-8
"""
Discord chat commands parser
"""

import argparse
import inspect
import sys
from types import GeneratorType

import argh
from argh import compat
from argh.constants import (ATTR_EXPECTS_NAMESPACE_OBJECT, ATTR_WRAPPED_EXCEPTIONS, ATTR_WRAPPED_EXCEPTIONS_PROCESSOR,
                            DEST_FUNCTION)
from argh.dispatching import ArghNamespace
from argh.exceptions import CommandError
from argh.utils import get_arg_spec

import esst.atis.chat_commands.atis_discord_commands
from esst import commands, core
from esst.discord_bot import abstract
from esst.discord_bot.chat_commands import dcs, esst_, mission, report, server

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _cancel_execution(*_):
    raise SystemExit(0)


def _get_function_from_namespace_obj(namespace_obj):
    if isinstance(namespace_obj, ArghNamespace):
        # our special namespace object keeps the stack of assigned functions
        try:
            func = namespace_obj.get_function()
        except (AttributeError, IndexError):
            return None
    else:
        # a custom (probably vanilla) namespace object keeps the last assigned
        # function; this may be wrong but at least something may work
        if not hasattr(namespace_obj, DEST_FUNCTION):
            return None
        func = getattr(namespace_obj, DEST_FUNCTION)

    if not func or not hasattr(func, '__call__'):
        return None

    return func


def _execute_command(func, namespace_obj, pre_call=None):  # noqa: C901
    # noinspection SpellCheckingInspection
    """
    Assumes that `function` is a callable.  Tries different approaches
    to call it (with `namespace_obj` or with ordinary signature).
    Yields the results line by line.

    If :class:`~argh.exceptions.CommandError` is raised, its message is
    appended to the results (i.e. yielded by the generator as a string).
    All other exceptions propagate unless marked as wrappable
    by :func:`wrap_errors`.
    """
    if pre_call:
        LOGGER.debug(f'running pre_call: {pre_call}')
        pre_call(namespace_obj)

    # namespace -> dictionary
    def _flat_key(key):
        return key.replace('-', '_')

    # noinspection SpellCheckingInspection
    def _call():
        # Actually call the function
        if getattr(func, ATTR_EXPECTS_NAMESPACE_OBJECT, False):
            result_ = func(namespace_obj)
        else:

            all_input = dict((_flat_key(k), v)
                             for k, v in vars(namespace_obj).items())

            # filter the namespace variables so that only those expected
            # by the actual function will pass

            spec = get_arg_spec(func)

            positional = [all_input[k] for k in spec.args]
            # noinspection SpellCheckingInspection
            kw_only = getattr(spec, 'kwonlyargs', [])
            keywords = dict((k, all_input[k]) for k in kw_only)

            # *args
            if spec.varargs:
                positional += getattr(namespace_obj, spec.varargs)

            # **kwargs
            varkw = getattr(spec, 'varkw', getattr(spec, 'keywords', []))
            if varkw:
                not_kwargs = [DEST_FUNCTION] + spec.args + [spec.varargs] + kw_only
                for k in vars(namespace_obj):
                    if k.startswith('_') or k in not_kwargs:
                        continue
                    keywords[k] = getattr(namespace_obj, k)

            result_ = func(*positional, **keywords)

        # Yield the results
        if isinstance(result_, (GeneratorType, list, tuple)):
            # yield each line ASAP, convert CommandError message to a line
            for line_ in result_:
                yield line_
        else:
            # yield non-empty non-iterable result as a single line
            if result_ is not None:
                yield result_

    # noinspection SpellCheckingInspection
    wrappable_exceptions = [CommandError, Exception]
    wrappable_exceptions += getattr(func, ATTR_WRAPPED_EXCEPTIONS, [])

    try:
        LOGGER.debug(f'running func {func}')
        result = _call()
        return '\n'.join(result)
    # pylint: disable=catching-non-exception
    except tuple(wrappable_exceptions) as exc:
        # pylint: disable=unnecessary-lambda
        processor = getattr(
            func, ATTR_WRAPPED_EXCEPTIONS_PROCESSOR,
            lambda exc_: '{0.__class__.__name__}: {0}'.format(exc_)
        )

        LOGGER.error(compat.text_type(processor(exc)))
        LOGGER.exception(exc)


class HelpFormatter(argparse.RawDescriptionHelpFormatter):
    """
    Dummy
    """

    def add_usage(self, usage, actions, groups, prefix=None):
        """
        Dummy

        Args:
            usage:
            actions:
            groups:
            prefix:

        Returns:

        """
        if usage and '!help' in usage:
            usage = ''
        return super(HelpFormatter, self).add_usage(usage, actions, groups, prefix)


# noinspection SpellCheckingInspection
class DiscordCommandParser(argh.ArghParser, abstract.AbstractDiscordCommandParser):
    """
    Creates chat commands out of regular functions with argh
    """

    def __init__(self,  # pylint: disable=too-many-arguments
                 prog=None,
                 usage=None,
                 description=None,
                 epilog=None,
                 parents=None,
                 formatter_class=HelpFormatter,
                 prefix_chars='-',
                 fromfile_prefix_chars=None,
                 argument_default=None,
                 conflict_handler='error',
                 add_help=True,
                 allow_abbrev=True):
        if parents is None:
            parents = []
        argh.ArghParser.__init__(
            self,
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            parents=parents,
            formatter_class=formatter_class,
            prefix_chars=prefix_chars,
            fromfile_prefix_chars=fromfile_prefix_chars,
            argument_default=argument_default,
            conflict_handler=conflict_handler,
            add_help=add_help,
            allow_abbrev=allow_abbrev,
        )
        self.formatter_class = HelpFormatter

    def _print_message(self, message, _=None):
        if message:
            commands.DISCORD.say(message)

    # pylint: disable=arguments-differ,too-many-branches,too-many-arguments,too-many-locals
    def dispatch(self,  # noqa: C901
                 argv=None,
                 add_help_command=True,
                 completion=True,  # pylint: disable=unused-argument
                 pre_call=None,
                 output_file=sys.stdout,  # pylint: disable=unused-argument
                 errors_file=sys.stderr,  # pylint: disable=unused-argument
                 raw_output=False,  # pylint: disable=unused-argument
                 namespace=None,
                 skip_unknown_args=False,
                 is_admin: bool = False, ):
        """
        Passes arguments to linked function

        Args:
            is_admin: is the user that issued the command an admin ?
            argv:
            add_help_command:
            completion:
            pre_call:
            output_file:
            errors_file:
            raw_output:
            namespace:
            skip_unknown_args:

        """
        try:
            for arg in argv:
                if arg == 'help':
                    argv.remove('help')
                    argv.append('--help')
            for arg in argv:
                if arg in ['-h', '--help']:
                    pre_call = _cancel_execution

            if argv is None:
                argv = sys.argv[1:]

            if add_help_command:
                if argv:
                    if argv[0] in ['help', '-h']:
                        argv.pop(0)
                        argv.append('--help')

            if skip_unknown_args:
                parse_args = self.parse_known_args
            else:
                parse_args = self.parse_args

            if not namespace:
                namespace = ArghNamespace()

            # this will raise SystemExit if parsing fails
            namespace_obj = parse_args(argv, namespace=namespace)

            func = _get_function_from_namespace_obj(namespace_obj)

            if func:
                if hasattr(func, 'protected_') and not is_admin:
                    LOGGER.error(
                        f'only users with role "{core.CFG.discord_admin_role}" have access to this command')
                    return None
                LOGGER.debug(f'running func: {func}')
                return _execute_command(func, namespace_obj, pre_call=pre_call)
            else:
                # no commands declared, can't dispatch; display help message
                return [self.format_usage()]

        except SystemExit:
            pass

    def exit(self, _=0, message=None):
        """
        This was supposed to exit the program in case of error; in this case, we simply print the message

        Args:
            _: exit code (useless)
            message: exit message

        """
        if message:
            self._print_message(message)

    def error(self, message):
        """error(message: string)

        Prints a usage message incorporating the message to stderr and
        exits.

        If you override this in a subclass, it should not return -- it
        should either exit or raise an exception.
        """
        self._print_message(message)
        raise SystemExit(-1)

    def parse_args(self, args=None, namespace=None):
        """
        Wrapper for :meth:`argparse.ArgumentParser.parse_args`.  If `namespace`
        is not defined, :class:`argh.dispatching.ArghNamespace` is used.
        This is required for functions to be properly used as commands.
        """
        if len(args) == 1:
            args.append('--help')
        try:
            return super(DiscordCommandParser, self).parse_args(args, namespace)
        except TypeError:
            pass

    def format_help(self):  # pylint: disable=useless-super-delegation
        """
        I'm not sure about this one, this is an experiment

        """
        return super(DiscordCommandParser, self).format_help()

    def parse_discord_message(self, message: str, is_admin: bool):
        """
        PArses message from Discord

        Args:
            message: message content
            is_admin: is sender an admin?

        """
        if message.startswith('!'):

            if message == '!help':
                self._print_message(self.format_help())
            else:
                formatted_args = message.split(' ')
                self.dispatch(formatted_args, is_admin=is_admin)


DESCRIPTION = """
To get help on a specific command, use the "--help" option.

Ex:
    !dcs --help
    !dcs status --help
"""
EPILOG = """
In addition to the commands listed above, you can also upload a mission to the server via Discord:
1. Drag and drop the mission file to this channel
2. In the "Add a comment" field, you can specify one of:
    a. "load": the mission will be loaded immediately (the server will restart)
    b. "overwrite": replace an existing mission file with the same name
    c. "force": force server restart even if there are connected players
    (note: those options can be combined: "load overwrite" is valid)
"""


def make_root_parser():
    """
    Creates the chat commands parser for the Discord bot

    Returns: parser object

    """
    parser = DiscordCommandParser(
        description=DESCRIPTION, prog='', add_help=False, usage='', epilog=EPILOG)
    for module_ in [esst_, mission, server, dcs, report, esst.atis.chat_commands.atis_discord_commands]:
        funcs = [o[1] for o in inspect.getmembers(module_, inspect.isfunction)
                 if o[1].__module__ == module_.__name__ and not o[1].__name__.startswith('_')]
        parser.add_commands(
            functions=funcs,
            namespace=module_.NAMESPACE,
            namespace_kwargs={
                'title': module_.TITLE,
            },
            func_kwargs={
                # 'parents': [parser],
            },
        )

    return parser



# coding=utf-8
"""
Manages commands related to ESST itself
"""
from esst import __version__, commands, core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def log():
    """
    Show ESST log file
    """
    commands.DISCORD.send_file('esst.log')


def changelog():
    """
    Show ESST changelog file
    """
    changelog_path = utils.get_esst_changelog_path()
    if changelog_path:
        commands.DISCORD.send_file(changelog_path)


def version():
    """
    Show ESST version
    """
    commands.DISCORD.say(f'ESST v{__version__}')


@utils.arg(protected=True)
def restart():
    """
    Restart ESST (protected)

    """
    if not core.CFG.restart:
        LOGGER.error('no restart command given in config')
    else:
        LOGGER.info('restarting ESST')
        core.CTX.restart = True
        core.CTX.exit = True


NAMESPACE = '!esst'
TITLE = 'Manage ESST application'



# coding=utf-8
"""
Manages config params for Univers Radio
"""

from elib.config import ConfigProp

NAMESPACE = 'ATIS'


# pylint: disable=too-few-public-methods
class ATISConfig:
    """
    Manages config params for Univers Radio
    """

    atis_create = ConfigProp(bool, default='true', namespace=NAMESPACE)



# coding=utf-8
"""
Globals
"""
import logging.handlers

import elib
# noinspection PyUnresolvedReferences
import emiz

from esst import __version__
from .context import Context
# from .config import Config
from .new_config import ESSTConfig, validate_config
from .new_config.setup_config import setup_config
from .status import ServerStatus, Status
from .fs_paths import FS

MAIN_LOGGER = elib.custom_logging.get_logger(
    logger_name='ESST',
    log_to_file='esst.log',
    rotate_logs=True,
)

MAIN_LOGGER.info(f'ESST version {__version__}')

EMIZ_LOGGER = logging.getLogger('EMIZ')
ELIB_LOGGER = logging.getLogger('ELIB')

for logger in (EMIZ_LOGGER, ELIB_LOGGER):
    for handler in logger.handlers:
        MAIN_LOGGER.debug(f'removing {handler} from {logger}')
        logger.removeHandler(handler)

for handler in MAIN_LOGGER.handlers:
    EMIZ_LOGGER.addHandler(handler)
    ELIB_LOGGER.addHandler(handler)

CTX = Context()
MAIN_LOGGER.info('reading config')
CFG = setup_config()



# coding=utf-8
"""
Install a handler to redirect all INFO messages (and higher) to the Discord Channel
"""

import logging

from esst.commands import DISCORD
from esst.core import MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


class DiscordLoggingHandler(logging.Handler):
    """
    Install a handler to redirect all INFO messages (and higher) to the Discord Channel
    """

    def __init__(self):
        logging.Handler.__init__(self, logging.INFO)

    def emit(self, record: logging.LogRecord):
        """
        Redirects the record to the Discord channel if its level is INFO or higher

        Args:
            record: logging.record to emit
        """
        if isinstance(record.msg, str):
            DISCORD.say(record.msg)


def register_logging_handler():
    """
    Installs the handler to the main logger
    """
    LOGGER.debug('registering Discord logging handler')
    MAIN_LOGGER.addHandler(DiscordLoggingHandler())



# coding=utf-8
"""
Tests for esst.dcs.mission_editor_lua
"""
import string
from pathlib import Path

import pytest
from hypothesis import given, settings
from hypothesis import strategies as st

from esst import core
from esst.dcs import mission_editor_lua

TEMPLATE = Path('./test/test_files/MissionEditor.lua').read_text(encoding='utf8')


@pytest.fixture(autouse=True)
def _setup():
    core.FS.dcs_path = Path('./DCS')
    core.FS.variant_saved_games_path = Path('./Saved Games')
    core.FS.mission_editor_lua_file = Path('./MissionEditor/MissionEditor.lua')
    Path('./DCS').mkdir()
    Path('./Saved Games').mkdir()


def test_injection():
    Path('./MissionEditor').mkdir()
    template_file = Path('./MissionEditor/MissionEditor.lua')
    template_file.write_text(TEMPLATE, encoding='utf8')
    assert mission_editor_lua.INJECT_TEMPLATE not in template_file.read_text(encoding='utf8')
    assert mission_editor_lua.inject_mission_editor_code()
    assert Path('./MissionEditor/MissionEditor.lua_backup_unknown').exists()
    content = template_file.read_text(encoding='utf8')
    assert mission_editor_lua.INJECT_TEMPLATE in content
    assert mission_editor_lua.inject_mission_editor_code()
    assert content == template_file.read_text(encoding='utf8')


def test_dcs_does_not_exist():
    with pytest.raises(FileNotFoundError):
        mission_editor_lua.inject_mission_editor_code()


def test_mission_editor_lua_does_not_exist():
    with pytest.raises(FileNotFoundError):
        mission_editor_lua.inject_mission_editor_code()


@given(text=st.text(min_size=20, max_size=200, alphabet=string.printable))
@settings(max_examples=1)
def test_wrong_content(text):
    Path('./MissionEditor').mkdir(exist_ok=True)
    template_file = Path('./MissionEditor/MissionEditor.lua')
    template_file.write_text(text, encoding='utf8')
    assert mission_editor_lua.INJECT_TEMPLATE not in template_file.read_text(encoding='utf8')
    assert not mission_editor_lua.inject_mission_editor_code()
    assert Path('./MissionEditor/MissionEditor.lua_backup_unknown').exists()
    assert mission_editor_lua.INJECT_TEMPLATE not in template_file.read_text(encoding='utf8')



# coding=utf-8
"""
Commands related to managing the server computer
"""

import humanize

from esst import commands, core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def status():
    """
    Show current server status
    """
    output = []
    for attr_name in dir(core.ServerStatus):
        if attr_name.startswith('_'):
            continue
        attr_nice_name = (attr_name[:1].upper() +
                          attr_name[1:]).replace("_", " ")
        value = getattr(core.ServerStatus, attr_name)
        if ('memory' in attr_name or 'swap' in attr_name) and value != 'unknown':
            value = humanize.naturalsize(value)
        if 'usage' in attr_name:
            value = str(value) + '%'
        output.append(f'{attr_nice_name}: {value}')
    commands.DISCORD.say('Server status:\n' + '\n'.join(output))


@utils.arg('--hours', help='Show stats for the last HOURS hours')
@utils.arg('--minutes', help='Show stats for the last MINUTES minutes')
@utils.arg('--days', help='Show stats for the last DAYS days')
def graph(days=0, hours=0, minutes=0):
    """
    Shows a graph of server performance (CPU, memory, ...)

    By default, the command shows the stats for the last 2 hours
    """
    if all((days == 0, hours == 0, minutes == 0)):
        hours = 2
    commands.SERVER.show_graph(days, hours, minutes)


@utils.arg('--start', help='Show CPU usage in real time')
@utils.arg('--stop', help='Stop showing CPU usage in real time')
def show_cpu(
        start=False,
        stop=False,
):
    """
    Show server CPU usage
    """
    if start:
        commands.SERVER.show_cpu_usage_start()
    elif stop:
        commands.SERVER.show_cpu_usage_stop()
    else:
        commands.SERVER.show_cpu_usage_once()


@utils.arg('--force', help='force server reboot, even when players are connected')
@utils.arg(protected=True)
def reboot(force: bool = False):
    """
    Restart the server computer (protected)
    """
    LOGGER.warning('rebooting server, ciao a tutti !')
    commands.SERVER.reboot(force)


def ip():  # pylint: disable=invalid-name
    """
    Show the server's external IP
    """
    commands.DISCORD.say(f'Server IP: {utils.external_ip()}')


NAMESPACE = '!server'
TITLE = 'Manage server computer'



# coding=utf-8
"""
Commands related to managing the DSC application
"""

import humanize

from esst import commands, core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def status():
    """
    Show current DCS status
    """
    output = []
    for attr_name in dir(core.Status):
        if attr_name.startswith('_'):
            continue
        attr_nice_name = attr_name[:1].upper() + attr_name[1:]
        attr_nice_name = attr_nice_name.replace("_", " ")
        if attr_name in ['mission_time', 'server_age']:
            output.append(f'{attr_nice_name}: '
                          f'{humanize.naturaltime(getattr(core.Status, attr_name))}')
        else:
            output.append(f'{attr_nice_name}: {getattr(core.Status, attr_name)}')
    commands.DISCORD.say('\n'.join(output))


@utils.arg('--start', help='Show CPU usage in real time')
@utils.arg('--stop', help='Stop showing CPU usage in real time')
def show_cpu(
        start=False,
        stop=False
):
    """
    Show DCS.exe CPU usage
    """
    if start:
        commands.DCS.show_cpu_usage_start()
    elif stop:
        commands.DCS.show_cpu_usage_stop()
    else:
        commands.DCS.show_cpu_usage_once()


@utils.arg('--force', help='force restart, even when players are connected')
@utils.arg(protected=True)
def restart(force: bool = False):
    """
    Closes and restart DCS.exe (protected)
    """
    commands.DCS.restart(force=force)


def version():
    """
    Show DCS.exe version
    """
    commands.DISCORD.say(f'DCS version: {core.Status.dcs_version}')


def log():
    """
    Show DCS log file
    """
    commands.DISCORD.send_file(utils.get_dcs_log_file_path())


NAMESPACE = '!dcs'
TITLE = 'Manage DCS application'



# coding=utf-8
"""
Catches connection error in Discord bot
"""

import asyncio

import aiohttp
import websockets.exceptions
from discord import HTTPException

from esst.core import CTX, MAIN_LOGGER

LOGGER = MAIN_LOGGER.getChild(__name__)


def _pass_exception(exc: Exception):
    LOGGER.error(exc)
    # LOGGER.exception('Discord bot error')
    # if CTX.sentry:
    #     CTX.sentry.captureException(exc)


def catch_exc(func):
    """
    Decorator to protect discord.client methods
    """

    async def _wrapper(*args, **kwargs):

        try:
            return await func(*args, **kwargs)
        except (websockets.exceptions.InvalidHandshake,
                websockets.exceptions.InvalidState,
                websockets.exceptions.PayloadTooBig,
                websockets.exceptions.WebSocketProtocolError,
                websockets.exceptions.InvalidURI,
                HTTPException) as exc:
            _pass_exception(exc)
        except (aiohttp.ClientError,
                ConnectionError,
                OSError,
                aiohttp.ClientOSError,
                aiohttp.ClientResponseError) as exc:
            await asyncio.sleep(10)
            while not CTX.wan:
                await asyncio.sleep(2)
            _pass_exception(exc)

    return _wrapper



# coding=utf-8
"""
Dummy class to facilitate the passing of information between other classes
"""


class Status:  # pylint: disable=too-few-public-methods
    """
    Represents DCS status
    """
    dcs_application = 'not running'
    dcs_version = 'unknown'
    dcs_cpu_usage = 'unknown'
    server_status = 'unknown'
    metar = 'unknown'
    mission_file = 'unknown'
    server_age = 'unknown'
    mission_time = 'unknown'
    paused = 'unknown'
    mission_name = 'unknown'
    players = []


class ServerStatus:  # pylint: disable=too-few-public-methods
    """
    Represents server status
    """
    physical_cpus = 'unknown'
    logical_cpus = 'unknown'
    cpu_frequency = 'unknown'
    cpu_usage = 'unknown'
    total_memory = 'unknown'
    used_memory = 'unknown'
    mem_usage = 'unknown'
    swap_size = 'unknown'
    swap_used = 'unknown'
    boot_time = 'unknown'
    bytes_recv_ = 0
    bytes_sent_ = 0
    bytes_recv = 'unknown'
    bytes_sent = 'unknown'



# coding=utf-8
"""
Tests for esst.mission.store
"""
import random
import string
from pathlib import Path

import pytest

from esst import core
from esst.mission import store


@pytest.fixture(autouse=True)
def _setup():
    core.FS.dcs_mission_folder = Path('.')


def _random_file_name(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for _ in range(size))


def test__get_mission_folder(tmpdir):
    path = store._get_mission_folder(str(tmpdir), 'ESST')
    assert isinstance(path, Path)
    assert path.exists()
    assert path.is_dir()


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test__get_mission_folder_path_exists(tmpdir, path):
    Path(str(tmpdir), path).mkdir()
    path = store._get_mission_folder(str(tmpdir), path)
    assert isinstance(path, Path)
    assert path.exists()
    assert path.is_dir()


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test__get_mission_folder_path_is_file(tmpdir, path):
    test_file = Path(str(tmpdir), path)
    with open(test_file, 'w') as stream:
        stream.write('test')
    with pytest.raises(RuntimeError):
        store._get_mission_folder(str(test_file))


def test_get_base_missions_folder():
    path = store.get_base_missions_folder()
    assert isinstance(path, Path)
    assert path.exists()
    assert path.is_dir()


def test_get_auto_missions_folder():
    path = store.get_auto_missions_folder()
    assert isinstance(path, Path)
    assert path.exists()
    assert path.is_dir()


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test_get_random_auto_mission_name(path):
    orig_miz_file = Path(store.get_base_missions_folder(), f'{path}.miz')
    orig_miz_file.touch()
    random_miz = store.get_random_auto_mission_name(orig_miz_file)
    assert random_miz.parent == store.get_auto_missions_folder()


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test_get_random_auto_mission_name_not_a_file(path):
    orig_miz_file = Path(store.get_base_missions_folder(), path)
    orig_miz_file.mkdir()
    with pytest.raises(RuntimeError):
        store.get_random_auto_mission_name(orig_miz_file)


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test_get_random_auto_mission_name_not_a_miz_file(path):
    orig_miz_file = Path(store.get_base_missions_folder(), f'{path}.not_miz')
    orig_miz_file.touch()
    with pytest.raises(RuntimeError):
        store.get_random_auto_mission_name(orig_miz_file)


@pytest.mark.parametrize('path', [_random_file_name() for _ in range(10)])
def test_get_random_auto_mission_name_does_not_exist(path):
    orig_miz_file = Path(store.get_base_missions_folder(), f'{path}')
    with pytest.raises(RuntimeError):
        store.get_random_auto_mission_name(orig_miz_file)


@pytest.mark.parametrize('_', [_random_file_name() for _ in range(20)])
def test_clean_store(_):
    list_of_files = [f'{_random_file_name()}.miz' for _ in range(20)]
    for file in list_of_files:
        p = Path(store.get_auto_missions_folder(), file)
        p.touch()
        assert p.is_file()
    for x in store.clean():
        assert x.name in list_of_files



# coding=utf-8
"""
ATIS Discord commands group
"""

from esst import core, utils

from .._atis_airfields import ALL_AIRFIELDS
from .._atis_get_info import get_info_for_icao
from .._univers_radio import URVoiceService

LOGGER = core.MAIN_LOGGER.getChild(__name__)


@utils.arg('icao', nargs=1)
def show(icao: list):
    """
    Show ATIS info for a specific airfield
    """
    icao = ''.join(icao).upper()
    try:
        info = get_info_for_icao(icao)
    except KeyError:
        LOGGER.error(f'ICAO not found in the list of currently active ATIS: {icao}')
        return
    running = 'running' if URVoiceService.is_running() else 'not running'
    info_str = f'UR voice service is {running}\n\n' \
               f'Metar: {core.Status.metar}\n' \
               f'Active runway: {info.active_runway}\n' \
               f'Information ID: {info.info_letter}\n' \
               f'ATIS speech: {core.CTX.atis_speech}'
    LOGGER.info(info_str)


def status():
    """
    Show UR voice service status
    """
    status_ = 'running' if URVoiceService.is_running() else 'not running'
    LOGGER.info(f'UniversRadio voice service is {status_}')


def frequencies():
    """
    Shows frequencies for the ATIS in Georgia
    """
    output = ['List of ATIS frequencies in Georgia:']
    for airfield in ALL_AIRFIELDS:
        output.append(f'{airfield.name}: {airfield.atis_freq.long_freq()}')
    LOGGER.info('\n'.join(output))


NAMESPACE = '!atis'
TITLE = 'Manage ATIS service'



# coding=utf-8
"""
Manages a UDP socket and does two things:

1. Retrieve incoming messages from DCS and update :py:class:`esst.core.status.status`
2. Sends command to the DCS application via the socket
"""

from .listener import DCSListener



# coding=utf-8
"""
Etcher's Server Startup Tool
"""
from pkg_resources import DistributionNotFound, get_distribution

try:
    __version__ = get_distribution('esst').version
except DistributionNotFound:  # pragma: no cover
    # package is not installed
    __version__ = 'not installed'



# coding=utf-8
"""
Manages DCS application Windows process
"""
import asyncio
import time

import psutil

from esst import commands, core, utils

from . import autoexec_cfg, mission_editor_lua, missions_manager
from .commands import DCS
from .dedicated import setup_config_for_dedicated_run
from .game_gui import install_game_gui_hooks
from .rotate_logs import rotate_dcs_log

LOGGER = core.MAIN_LOGGER.getChild(__name__)


async def get_dcs_process_pid():
    """
    Tries to obtain "DCS.exe" PID

    Returns: "DCS.exe" PID or False
    """
    for process in psutil.process_iter():
        try:
            if process.name().lower() == 'dcs.exe':
                return process.pid
        except psutil.NoSuchProcess:
            pass

    return False


# pylint: disable=too-few-public-methods,too-many-instance-attributes
class App:
    """
    Manages DCS application Window process
    """
    # noinspection SpellCheckingInspection
    valid_priorities = {
        'idle': psutil.IDLE_PRIORITY_CLASS,
        'below_normal': psutil.BELOW_NORMAL_PRIORITY_CLASS,
        'normal': psutil.NORMAL_PRIORITY_CLASS,
        'above_normal': psutil.ABOVE_NORMAL_PRIORITY_CLASS,
        'high': psutil.HIGH_PRIORITY_CLASS,
        'realtime': psutil.REALTIME_PRIORITY_CLASS,
    }

    def __init__(self):
        self._app = None
        self.process_pid = None
        self._restart_ok = True
        self._blockers_warned = set()
        self._warned = False
        # self._additional_parameters = []

    @property
    def app(self) -> psutil.Process:
        """Process instance"""
        return self._app

    # noinspection PyMethodMayBeStatic
    async def _execute_cmd_chain(self, cmd_chain: list):
        while not core.CTX.exit:
            try:
                cmd = cmd_chain.pop(0)
                await cmd()
                await asyncio.sleep(0.1)
            except IndexError:
                return

    # noinspection PyMethodMayBeStatic
    async def _get_dcs_version_from_executable(self):
        # noinspection PyBroadException
        core.Status.dcs_version = utils.Win32FileInfo(str(core.FS.dcs_exe)).file_version
        LOGGER.debug(f'DCS version: {core.Status.dcs_version}')
        simplified_version = int(''.join(core.Status.dcs_version.split('.')[:3]))
        LOGGER.debug(f'simplified version: {simplified_version}')
        if simplified_version <= 157:
            pass
        elif simplified_version >= 158:
            mission_editor_lua.inject_mission_editor_code()
            autoexec_cfg.inject_silent_crash_report()
        setup_config_for_dedicated_run()
        return True

    async def _check_if_dcs_is_running(self):
        self.process_pid = await get_dcs_process_pid()
        if self.process_pid:
            await self._update_application_status('running')
        else:
            await self._update_application_status('not running')

    async def _try_to_connect_to_existing_dcs_application(self):
        if self.app and self.app.is_running():
            return
        LOGGER.debug('connecting to existing DCS application')
        await self._check_if_dcs_is_running()
        if self.process_pid:
            self._app = psutil.Process(self.process_pid)
            await self._wait_for_dcs_to_start()

    async def _wait_for_dcs_to_start(self):

        async def _wait_for_process():
            while True:
                if core.CTX.exit:
                    return
                await asyncio.sleep(0.1)
                if self.app.is_running():
                    break

        LOGGER.debug('waiting for DCS to spool up')
        await _wait_for_process()
        LOGGER.debug('process is ready')

    def _work_with_dcs_process(self, func):
        if core.CTX.exit:
            return
        try:
            func()
            self._warned = False
        except (psutil.NoSuchProcess, AttributeError):
            if not core.CTX.exit and not self._warned:
                LOGGER.warning('DCS process does not exist')
                self._warned = True

    def set_affinity(self):
        """
        Sets the DCS process CPU affinity to the CFG value
        """
        def _command():
            if list(self._app.cpu_affinity()) != list(core.CFG.dcs_cpu_affinity):
                LOGGER.debug(f'setting DCS process affinity to: {core.CFG.dcs_cpu_affinity}')
                self._app.cpu_affinity(list(core.CFG.dcs_cpu_affinity))

        while True:
            if core.CFG.dcs_cpu_affinity:
                if core.CTX.exit:
                    return
                self._work_with_dcs_process(_command)
            else:
                LOGGER.warning('no affinity given in config file')
                return
            time.sleep(30)

    def set_priority(self):
        """
        Sets the DCS process CPU priority to the CFG value
        """
        def _command():
            if self.app.nice() != self.valid_priorities[core.CFG.dcs_cpu_priority]:
                LOGGER.debug(
                    f'setting DCS process priority to: {core.CFG.dcs_cpu_priority}')
                self.app.nice(self.valid_priorities[core.CFG.dcs_cpu_priority])
        time.sleep(15)
        while True:
            if core.CFG.dcs_cpu_priority:
                if core.CTX.exit:
                    return
                if core.CFG.dcs_cpu_priority not in self.valid_priorities.keys():
                    LOGGER.error(f'invalid priority: {core.CFG.dcs_cpu_priority}\n'
                                 f'Choose one of: {self.valid_priorities.keys()}')
                    return
                self._work_with_dcs_process(_command)
            else:
                LOGGER.warning('no priority given in config file')
                return
            time.sleep(30)

    async def _start_new_dcs_application_if_needed(self):

        async def _start_dcs_app():
            LOGGER.debug(f'starting DCS application process: {core.FS.dcs_exe}')
            self._app = psutil.Popen(str(core.FS.dcs_exe))

        if self.app and self.app.is_running():
            return
        LOGGER.debug('starting new DCS application')
        cmd_chain = [
            _start_dcs_app,
            self._wait_for_dcs_to_start,
        ]
        rotate_dcs_log()
        await self._execute_cmd_chain(cmd_chain)
        await self._check_if_dcs_is_running()

    # noinspection PyMethodMayBeStatic
    async def _update_application_status(self, status: str):
        if core.Status.dcs_application != status:
            core.Status.dcs_application = status
            LOGGER.info(f'DCS application is {status}')
            if status == 'starting':
                commands.LISTENER.monitor_server_startup_start()

    async def kill_running_app(self):  # noqa: C901
        """
        Kills the running DCS.exe process
        """

        async def _ask_politely():
            if not self.app or not self.app.is_running():
                return True
            LOGGER.debug('sending socket command to DCS for graceful exit')
            commands.LISTENER.exit_dcs()
            await asyncio.sleep(1)
            LOGGER.debug(
                f'waiting on DCS to close itself (grace period: {core.CFG.dcs_grace_period})')
            now_ = utils.now()
            while self.app.is_running():
                await asyncio.sleep(1)
                if utils.now() - now_ > core.CFG.dcs_grace_period:
                    LOGGER.debug('grace period time out!')
                    return False

            LOGGER.info('DCS closed itself, nice')
            return True

        async def _no_more_mr_nice_guy():
            if not self.app or not self.app.is_running():
                return True
            LOGGER.debug('killing dcs.exe application')
            self.app.kill()
            now_ = utils.now()
            while self.app.is_running():
                await asyncio.sleep(1)
                if utils.now() - now_ > 10:
                    return False

            return True

        core.CTX.dcs_do_kill = False
        await self._check_if_dcs_is_running()
        if not self.app or not self.app.is_running():
            LOGGER.debug('DCS process was not running')
            return
        LOGGER.info('closing DCS')
        if not await _ask_politely():
            LOGGER.info('DCS will not exit gracefully, killing it')
            if not await _no_more_mr_nice_guy():
                LOGGER.error('I was not able to kill DCS, something is wrong')
                raise RuntimeError()
        await self._check_if_dcs_is_running()

    async def restart(self):
        """
        Restarts DCS application
        """
        if core.CTX.exit:
            LOGGER.debug('restart interrupted by exit signal')
            return
        core.CTX.dcs_do_restart = False
        LOGGER.info('restarting DCS')
        await self.kill_running_app()
        core.Status.mission_file = 'unknown'
        core.Status.server_age = 'unknown'
        core.Status.mission_time = 'unknown'
        core.Status.paused = 'unknown'
        core.Status.mission_name = 'unknown'
        core.Status.players = []
        self._app = None
        self.process_pid = None
        cmd_chain = [
            self._start_new_dcs_application_if_needed,
        ]
        await self._execute_cmd_chain(cmd_chain)

    def monitor_cpu_usage(self):
        """
        Gets the CPU usage of "DCS.exe" over 5 seconds, and sends an alert if the given threshold is exceeded

        Threshold is set via the config value "DCS_HIGH_CPU_USAGE", and it defaults to 80%
        """
        while not core.CTX.exit:
            try:
                if self.app and self.app.is_running():
                    cpu_usage = int(self.app.cpu_percent(core.CFG.dcs_high_cpu_usage_interval))
                    mem_usage = int(self.app.memory_percent())
                    core.Status.dcs_cpu_usage = f'{cpu_usage}%'
                    if core.CTX.dcs_show_cpu_usage or core.CTX.dcs_show_cpu_usage_once:
                        commands.DISCORD.say(f'DCS cpu usage: {cpu_usage}%')
                        core.CTX.dcs_show_cpu_usage_once = False
                    if core.CFG.dcs_high_cpu_usage:
                        if cpu_usage > core.CFG.dcs_high_cpu_usage and not core.Status.paused:
                            LOGGER.warning(
                                f'DCS cpu usage has been higher than {core.CFG.dcs_high_cpu_usage}%'
                                f' for {core.CFG.dcs_high_cpu_usage_interval} seconds')

                    now_ = utils.now()
                    core.CTX.dcs_mem_history.append((now_, mem_usage))
                    core.CTX.dcs_cpu_history.append((now_, cpu_usage))

            except psutil.NoSuchProcess:
                pass

            # I didn't think it could, happen, but of course it did ...
            # See https://github.com/132nd-vWing/ESST/issues/59
            except AttributeError:
                pass

    # pylint: disable=too-many-branches
    async def run(self):  # noqa: C901
        """
        Entry point of the thread
        """
        if not core.CTX.start_dcs_loop:
            LOGGER.debug('skipping DCS application loop')
            return
        if not await self._get_dcs_version_from_executable():
            return
        await core.CTX.loop.run_in_executor(None, install_game_gui_hooks)
        await core.CTX.loop.run_in_executor(None, missions_manager.get_latest_mission_from_github)
        await core.CTX.loop.run_in_executor(None, missions_manager.initial_setup)

        LOGGER.debug('starting DCS monitoring thread')
        if DCS.dcs_cannot_start():
            blockers = ", ".join(DCS.dcs_cannot_start())
            if blockers not in self._blockers_warned:
                self._blockers_warned.add(blockers)
                LOGGER.warning(f'DCS is prevented to start by: {", ".join(DCS.dcs_cannot_start())}')
        else:
            if self._blockers_warned:
                self._blockers_warned = set()
            await self._try_to_connect_to_existing_dcs_application()
            await self._start_new_dcs_application_if_needed()
        cpu_monitor_thread = core.CTX.loop.run_in_executor(None, self.monitor_cpu_usage)
        cpu_affinity_thread = core.CTX.loop.run_in_executor(None, self.set_affinity)
        cpu_priority_thread = core.CTX.loop.run_in_executor(None, self.set_priority)
        while True:
            if core.CTX.exit:
                LOGGER.debug('interrupted by exit signal')
                await cpu_monitor_thread
                await cpu_affinity_thread
                await cpu_priority_thread
                break
            if DCS.dcs_cannot_start():
                blockers = ", ".join(DCS.dcs_cannot_start())
                if blockers not in self._blockers_warned:
                    self._blockers_warned.add(blockers)
                    LOGGER.warning(f'DCS is prevented to start by: {", ".join(DCS.dcs_cannot_start())}')
            else:
                if self._blockers_warned:
                    self._blockers_warned = set()
                await self._check_if_dcs_is_running()
                if not self.process_pid:
                    LOGGER.debug('DCS has stopped, re-starting')
                    await self.restart()
            if core.CTX.dcs_do_kill:
                await self.kill_running_app()
            if core.CTX.dcs_do_restart:
                await self.restart()
            await asyncio.sleep(0.1)

        LOGGER.debug('end of DCS loop')



# coding=utf-8
"""
Creates graphic of perfs
"""

import datetime
import typing
from collections import namedtuple
from tempfile import mktemp

import humanize

from esst.core import CTX

PLT = GRID_SPEC = TICKER = None


# https://stackoverflow.com/questions/4931376/generating-matplotlib-graphs-without-a-running-x-server/4935945#4935945
# noinspection SpellCheckingInspection
def _init_mpl():
    """
    This is a very stupid hack to go around Matplotlib being stupid about Tkinter.
    My linters don't like import statements mixed within the code, so this will do.
    """
    global PLT, GRID_SPEC, TICKER  # pylint: disable=global-statement
    import matplotlib as mpl
    mpl.use('Agg')
    from matplotlib import pyplot as plt_
    from matplotlib import gridspec as grd_, ticker as tick_
    PLT = plt_
    GRID_SPEC = grd_
    TICKER = tick_


_init_mpl()

GraphValues = namedtuple('GraphValues', ['server_cpu_history',
                                         'server_mem_history',
                                         'server_bytes_sent_history',
                                         'server_bytes_recv_history',
                                         'dcs_cpu_history',
                                         'dcs_mem_history',
                                         'players_history', ])

PlotLine = namedtuple('PlotValue',
                      [
                          'values',
                          'label',
                          'style',
                      ])


def process_values(values_to_process: GraphValues, time_delta: float) -> GraphValues:
    """
    Converts raw values for plotting

    Args:
        values_to_process: values in set from CTX
        time_delta: how far behind?

    Returns: processed values

    """

    def _process(values):
        return [data for data in values if data[0] >= time_delta] or [(time_delta, 0)]

    server_cpu_history = _process(values_to_process.server_cpu_history)
    server_mem_history = _process(values_to_process.server_mem_history)
    server_bytes_sent_history = _process(values_to_process.server_bytes_sent_history)
    server_bytes_recv_history = _process(values_to_process.server_bytes_recv_history)
    dcs_cpu_history = _process(values_to_process.dcs_cpu_history)
    dcs_mem_history = _process(values_to_process.dcs_mem_history)
    players_history = _process(values_to_process.players_history)

    return GraphValues(
        server_cpu_history=zip(*server_cpu_history),
        server_mem_history=zip(*server_mem_history),
        server_bytes_sent_history=zip(*server_bytes_sent_history),
        server_bytes_recv_history=zip(*server_bytes_recv_history),
        dcs_cpu_history=zip(*dcs_cpu_history),
        dcs_mem_history=zip(*dcs_mem_history),
        players_history=tuple(zip(*players_history)),
    )


def _make_delta(now, days, hours, minutes):
    delta = datetime.timedelta(
        days=days, hours=hours, minutes=minutes).total_seconds()
    if delta == 0:
        delta = datetime.timedelta(hours=2).total_seconds()
    return now - delta


def _x_format_func(val, _):
    val = datetime.datetime.fromtimestamp(val)
    return str(val).replace(' ', '\n')


def _y_format_func_percent(val, _):
    return str(int(val)) + '%'


def _y_format_func_bytes(val, _):
    return humanize.naturalsize(val)


def _plot_axis(grid_spec, grid_pos,  # pylint: disable=too-many-arguments
               values_to_plot: typing.Set[PlotLine],
               title,
               y_label_text,
               values,
               now,
               y_format_func,
               visible_x_labels=False,
               share_x=None):
    axis = PLT.subplot(grid_spec[grid_pos], sharex=share_x)
    axis.set_title(title)
    PLT.setp(axis.get_xticklabels(), visible=visible_x_labels)
    axis.set_ylabel(y_label_text)

    for line in values_to_plot:
        assert isinstance(line, PlotLine)
        line_, = axis.plot(*line.values, line.style)
        PLT.setp(line_, label=line.label)
    _add_players_count_to_axis(axis, values.players_history)
    axis.xaxis.set_major_formatter(TICKER.FuncFormatter(_x_format_func))
    axis.yaxis.set_major_formatter(TICKER.FuncFormatter(y_format_func))
    axis.grid(True)
    axis.set_xlim(right=now)

    return axis


# pylint: disable=too-many-arguments,too-many-locals
def _get_axis(
        grid_spec,
        now,
        values,
        grid_pos,
        values_list: typing.List[typing.Any],
        labels_list: typing.List[str],
        title: str,
        y_label: str,
        visible_x: bool,
        y_format_func: callable,
        share_x=None,
):
    lines_to_plot = set()
    styles = ['r', 'b']
    for _values, _label in zip(values_list, labels_list):
        lines_to_plot.add(
            PlotLine(
                values=_values,
                style=styles.pop(),
                label=_label
            )
        )
    axis = _plot_axis(grid_spec,
                      now=now,
                      values_to_plot=lines_to_plot,
                      grid_pos=grid_pos,
                      title=title,
                      y_label_text=y_label,
                      values=values,
                      visible_x_labels=visible_x,
                      share_x=share_x,
                      y_format_func=y_format_func)
    return axis


def _plot_server(grid_spec, values, now):
    axis = _get_axis(
        grid_spec=grid_spec,
        now=now,
        values=values,
        grid_pos=0,
        values_list=[values.server_cpu_history, values.server_mem_history],
        labels_list=['CPU', 'Memory'],
        title='Server stats',
        y_label='Percentage used',
        visible_x=False,
        y_format_func=_y_format_func_percent,
    )
    axis.set_ylim([0, 100])
    return axis


def _plot_dcs(grid_spec, values, now, share_x=None):
    axis = _get_axis(
        grid_spec=grid_spec,
        now=now,
        values=values,
        grid_pos=1,
        values_list=[values.dcs_cpu_history, values.dcs_mem_history],
        labels_list=['CPU', 'Memory'],
        title='DCS stats',
        y_label='Percentage used',
        visible_x=False,
        y_format_func=_y_format_func_percent,
        share_x=share_x
    )
    axis.set_ylim([0, 100])
    return axis


def _plot_bandwidth(grid_spec, values, now, share_x=None):
    axis = _get_axis(
        grid_spec=grid_spec,
        now=now,
        values=values,
        grid_pos=2,
        values_list=[values.server_bytes_sent_history, values.server_bytes_recv_history],
        labels_list=['Bytes sent', 'Bytes received'],
        title='Bytes sent',
        y_label='Bytes received',
        visible_x=True,
        y_format_func=_y_format_func_bytes,
        share_x=share_x
    )
    return axis


def _add_players_count_to_axis(axis, players_history):
    ax_players = axis.twinx()
    max_player_count = max(
        max((players_count for players_count in players_history[1])), 10)
    ax_players.set_ylim([0, max_player_count + (max_player_count / 4)])
    ax_players.yaxis.set_major_locator(TICKER.MaxNLocator(integer=True))
    ax_players.set_ylabel('Connected players')
    players_history, = ax_players.plot(*players_history, 'k.', )
    PLT.setp(players_history, label='Players count')

    lines, labels = axis.get_legend_handles_labels()
    lines2, labels2 = ax_players.get_legend_handles_labels()
    axis.legend(lines + lines2, labels + labels2)


def _make_history_graph(  # pylint: disable=too-many-arguments
        values_to_process,
        days=0,
        hours=0,
        minutes=0,
        show: bool = False,
        save_path=None):
    """
    Creates a graph of perfs

    Args:
        show: show and exit
        save_path: specify path to save to (default to temp path)

    """
    # noinspection PyTypeChecker
    now = datetime.datetime.now().timestamp()
    time_delta = _make_delta(now, days, hours, minutes)

    values = process_values(values_to_process, time_delta)

    figure = PLT.figure(figsize=(18, 12))
    grid_spec = GRID_SPEC.GridSpec(3, 1, height_ratios=[1, 1, 1])

    ax_server = _plot_server(grid_spec, values, now)
    _plot_dcs(grid_spec, values, now, share_x=ax_server)
    _plot_bandwidth(grid_spec, values, now, share_x=ax_server)

    PLT.tight_layout()
    figure.tight_layout()

    if show:
        PLT.show()
        PLT.close()
        return None
    else:
        if not save_path:
            save_path = mktemp('.png')
        PLT.savefig(save_path)
        return save_path


# pylint: disable=too-many-arguments
def make_history_graph(callback=None, days=0, hours=0, minutes=0, show: bool = False, save_path=None):
    """
    Creates a graph of perfs

    Args:
        minutes: number of minutes to go back
        hours: number of hours to go back
        days: number of days to go back
        callback: optional call back to the future
        show: show and exit
        save_path: specify path to save to (default to temp path)

    """

    values_to_process = GraphValues(
        dcs_cpu_history=CTX.dcs_cpu_history,
        dcs_mem_history=CTX.dcs_mem_history,
        server_cpu_history=CTX.server_cpu_history,
        server_mem_history=CTX.server_mem_history,
        server_bytes_recv_history=CTX.server_bytes_recv_history,
        server_bytes_sent_history=CTX.server_bytes_sent_history,
        players_history=CTX.players_history,
    )
    future = CTX.process_pool.submit(
        _make_history_graph, values_to_process, days, hours, minutes, show, save_path)
    if callback:
        future.add_done_callback(callback)


if __name__ == '__main__':
    # Debug code
    import random

    TIME_DELTA = datetime.timedelta(hours=5)
    TOTAL_SECONDS = int(TIME_DELTA.total_seconds())

    NOW = datetime.datetime.now().timestamp()

    PLAYER_COUNT = 0
    CTX.players_history.append((NOW - TOTAL_SECONDS, 0))
    SKIP = 0
    for time_stamp in range(TOTAL_SECONDS, 0, -10):
        CTX.server_mem_history.append(
            (NOW - time_stamp, random.randint(60, 70)))
        CTX.dcs_cpu_history.append((NOW - time_stamp, random.randint(20, 30)))
        CTX.dcs_mem_history.append((NOW - time_stamp, random.randint(60, 70)))

        SKIP += 1
        if SKIP > 20:
            SKIP = 0
            CTX.server_bytes_recv_history.append(
                (NOW - time_stamp, random.randint(0, 50000000)))
            CTX.server_bytes_sent_history.append(
                (NOW - time_stamp, random.randint(0, 50000000)))

        if time_stamp <= int(TOTAL_SECONDS / 2):
            CTX.server_cpu_history.append(
                (NOW - time_stamp, random.randint(20, 30)))
        if random.randint(0, 100) > 99:
            PLAYER_COUNT += random.choice([-1, 1])
            if PLAYER_COUNT < 0:
                PLAYER_COUNT = 0
                continue
        CTX.players_history.append((NOW - time_stamp, PLAYER_COUNT))

    TIME_DELTA = datetime.datetime.now() - TIME_DELTA
    TIME_DELTA = TIME_DELTA.timestamp()

    make_history_graph(hours=5, save_path='./test.png')



# coding=utf-8
"""
Manages all FileSystem path for ESST
"""
import logging
import typing
from pathlib import Path

import elib

from esst.core.new_config import ESSTConfig

try:
    import winreg
except ImportError:
    from unittest.mock import MagicMock

    winreg = MagicMock()

LOGGER = logging.getLogger('ESST').getChild(__name__)

A_REG = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)


class FS:
    """
    Manages all FileSystem path for ESST
    """

    dcs_path: str = None
    dcs_exe: Path = None
    dcs_autoexec_file: Path = None
    dcs_hook_path: Path = None
    dcs_mission_folder: Path = None
    dcs_server_settings: Path = None
    dcs_logs_dir: Path = None

    mission_editor_lua_file: Path = None

    saved_games_path: str = None
    variant_saved_games_path: str = None

    ur_settings_folder: str = None
    ur_voice_settings_file: str = None
    ur_install_path: str = None

    @staticmethod
    def _reset():
        """Testing only"""
        FS.dcs_path = None
        FS.dcs_exe = None
        FS.dcs_autoexec_file = None
        FS.mission_editor_lua_file = None
        FS.saved_games_path = None
        FS.variant_saved_games_path = None
        FS.ur_install_path = None
        FS.ur_settings_folder = None
        FS.ur_install_path = None

    @staticmethod
    def ensure_path(path: typing.Union[str, Path], path_name: str, must_exist=True) -> Path:
        """
        Makes sure that "path" is a Path instance, and (optionally) exists

        Args:
            path: str or Path to check
            path_name: human friendly description of the path
            must_exist: raises FileNotFoundError if True and path does not exist

        Returns: Path instance

        """
        if path is None:
            raise RuntimeError(f'path uninitialized: {path_name}')
        return elib.path.ensure_path(path, must_exist=must_exist)

    @staticmethod
    def get_saved_games_variant(dcs_path: typing.Union[str, Path] = None) -> Path:
        """
        Infers Saved Games dir specific to this DCS installation by reading the (optional) "dcs_variant.txt"
        file that is contained at the root of the DCS installation

        Args:
            dcs_path:  path to the DCS installation

        Returns: Path instance for the Saved Games/DCS[variant] dir

        """
        if dcs_path is None:
            dcs_path = FS.ensure_path(FS.dcs_path, 'dcs path')
        FS.ensure_path(FS.saved_games_path, 'saved games')

        dcs_path = elib.path.ensure_dir(dcs_path)
        variant_path = Path(dcs_path, 'dcs_variant.txt')
        if variant_path.exists():
            variant = f'.{variant_path.read_text(encoding="utf8")}'
        else:
            variant = ''
        return elib.path.ensure_dir(FS.saved_games_path, f'DCS{variant}').absolute()

    @staticmethod
    def _get_saved_games_from_registry() -> Path:
        LOGGER.debug('searching for base "Saved Games" folder')
        try:
            LOGGER.debug('trying "User Shell Folders"')
            with winreg.OpenKey(A_REG, r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\User Shell Folders") as key:
                # noinspection SpellCheckingInspection
                base_sg = Path(winreg.QueryValueEx(key, "{4C5C32FF-BB9D-43B0-B5B4-2D72E54EAAA4}")[0])
                LOGGER.debug('found in "User Shell Folders"')
        except FileNotFoundError:
            LOGGER.debug('failed, trying "Shell Folders"')
            try:
                with winreg.OpenKey(A_REG, r"SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders") as key:
                    # noinspection SpellCheckingInspection
                    base_sg = Path(winreg.QueryValueEx(key, "{4C5C32FF-BB9D-43B0-B5B4-2D72E54EAAA4}")[0])
                    LOGGER.debug('found in "Shell Folders"')
            except FileNotFoundError:
                LOGGER.debug('darn it, another fail, falling back to "~"')
                base_sg = Path('~').expanduser().abspath()
        return base_sg

    @staticmethod
    def discover_saved_games_path(cfg: ESSTConfig):
        """
        Tries to find Saved Games on this system

        Returns: Saved Games dir
        """
        if not cfg.saved_games_dir:
            LOGGER.debug('no Saved Games path in Config, looking it up')
            return FS._get_saved_games_from_registry()

        LOGGER.debug('Saved Games path found in Config')
        base_sg = Path(cfg.saved_games_dir)
        if not base_sg.is_dir():
            LOGGER.error(f'Saved Games dir provided in config file is invalid: {base_sg}')
            return FS._get_saved_games_from_registry()

        return base_sg


def _init_saved_games(cfg):
    FS.saved_games_path = FS.discover_saved_games_path(cfg)
    LOGGER.debug(f'Saved Games path: {FS.saved_games_path}')


def _init_dcs_path(cfg):
    FS.dcs_path = elib.path.ensure_dir(
        cfg.dcs_path
    ).absolute()
    LOGGER.debug(f'DCS path: {FS.dcs_path}')


def _init_dcs_exe():
    FS.dcs_exe = elib.path.ensure_file(
        FS.dcs_path,
        'bin/dcs.exe'
    ).absolute()
    LOGGER.debug(f'DCS exe: {FS.dcs_exe}')


def _init_saved_games_variant():
    FS.variant_saved_games_path = FS.get_saved_games_variant(
        FS.dcs_path
    ).absolute()
    LOGGER.debug(f'Saved Games variant: {FS.variant_saved_games_path}')


def _init_autoexec_cfg():
    FS.dcs_autoexec_file = Path(
        FS.variant_saved_games_path,
        'Config/autoexec.cfg'
    ).absolute()
    LOGGER.debug(f'DCS autoexec: {FS.dcs_autoexec_file}')


def _init_mission_editor_lua():
    FS.mission_editor_lua_file = elib.path.ensure_file(
        FS.dcs_path,
        'MissionEditor/MissionEditor.lua'
    ).absolute()
    LOGGER.debug(f'Mission Editor lua file: {FS.mission_editor_lua_file}')


def _init_hooks():
    FS.dcs_hook_path = elib.path.ensure_file(
        FS.variant_saved_games_path,
        'Scripts/Hooks/esst.lua',
        must_exist=False
    ).absolute()
    LOGGER.debug(f'DCS hook: {FS.dcs_hook_path}')


def _init_mission_folder():
    FS.dcs_mission_folder = elib.path.ensure_dir(
        FS.variant_saved_games_path,
        'Missions/ESST',
        must_exist=False,
        create=True
    ).absolute()
    LOGGER.debug(f'DCS mission folder: {FS.dcs_mission_folder}')


def _init_server_settings():
    FS.dcs_server_settings = elib.path.ensure_file(
        FS.variant_saved_games_path,
        'Config/serverSettings.lua',
        must_exist=False
    ).absolute()
    LOGGER.debug(f'DCS server settings: {FS.dcs_server_settings}')


def _init_logs_folder():
    FS.dcs_logs_dir = elib.path.ensure_path(
        FS.variant_saved_games_path,
        'logs',
        must_exist=False,
    ).absolute()
    LOGGER.debug(f'DCS log folder: {FS.dcs_logs_dir}')


def init_fs(cfg: ESSTConfig):
    """
    Initializes File System paths

    Args:
        cfg: ESST configuration

    """
    LOGGER.debug('init')
    _init_saved_games(cfg)
    _init_dcs_path(cfg)
    _init_dcs_exe()
    _init_saved_games_variant()
    _init_autoexec_cfg()
    _init_mission_editor_lua()
    _init_hooks()
    _init_mission_folder()
    _init_server_settings()
    _init_logs_folder()
    LOGGER.debug('FS paths initialised')



# coding=utf-8
"""
Manages all things related to UR
"""

from ._ur_voice_service_settings import URVoiceServiceSettings
from ._ur_voice_service import URVoiceService
from ._ur_install_dir import discover_ur_install_path
from ._ur_object import Airfield, URCoord, URFrequency
from ._ur_status import Status



# coding=utf-8
"""
UniversRadio status
"""
import esst.utils.sentry_context


class Status(esst.utils.sentry_context.SentryClassContext):
    """
    UniversRadio status
    """

    def __init__(self):
        pass

    install_path = 'unknown'
    settings_folder = 'unknown'
    voice_settings_file = 'unknown'



# coding=utf-8
"""
Meh
"""
import uuid

from esst import commands, core, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _send(msg, ctx):
    if msg and core.CTX.sentry:
        core.CTX.sentry.captureMessage(
            ctx + '_' + str(uuid.uuid4()), data={'extra': {'text': ' '.join(msg)}})
        commands.DISCORD.say('Thank you !')


@utils.arg('message', nargs='+', metavar='ISSUE')
def issue(message):
    """
    Send feedback about an issue with ESST
    """
    _send(message, 'ISSUE')


@utils.arg('message', nargs='+', metavar='SUGGESTION')
def suggestion(message):
    """
    Suggest a new feature
    """
    _send(message, 'SUGGESTION')


NAMESPACE = '!report'
TITLE = 'Report issues, request features, ...'



# coding=utf-8
"""
Main entry point
"""
import asyncio
import os
import queue
from pathlib import Path

import click

from esst import __version__
from esst.core import CFG, CTX, MAIN_LOGGER, fs_paths

MAIN_LOGGER.debug(f'Starting ESST version {__version__}')


async def watch_for_exceptions():
    """
    Dummy loop to wake up asyncio event loop from time to time
    """
    while True:
        if CTX.exit:
            break
        await asyncio.sleep(0.1)


# pylint: disable=too-many-locals,too-many-arguments
@click.group(invoke_without_command=True)  # noqa: C901
@click.option('--discord/--no-discord', default=True, help='Starts the Discord bot loop', show_default=True)
@click.option('--server/--no-server', default=True, help='Starts the server monitoring loop', show_default=True)
@click.option('--dcs/--no-dcs', default=True, help='Starts the DCS app loop', show_default=True)
@click.option('--listener/--no-listener', default=True, help='Starts the socket loop', show_default=True)
@click.option('--start-dcs/--no-start-dcs', help='Spawn DCS.exe process', default=True, show_default=True)
@click.option('--install-hooks/--no-install-hooks', help='Install GameGUI hooks', default=True, show_default=True)
@click.option('--install-dedi-config/--no-install-dedi-config', help='Setup DCS to run in dedicated mode', default=True,
              show_default=True)
@click.option('--auto-mission/--no-auto-mission', help='Download latest mission', default=True, show_default=True)
def main(
        discord: bool,
        server: bool,
        dcs: bool,
        listener: bool,
        start_dcs: bool,
        install_hooks: bool,
        install_dedi_config: bool,
        auto_mission: bool,):
    """
    Main entry point

    Args:
        install_dedi_config: setup DCS to run in dedicated mode
        install_hooks: install GameGUI hooks
        dcs: start dcs loop
        discord: start Discord bot loop
        server: start server loop
        listener: start the listener loop
        start_dcs: start the server thread, but not the actual DCS app
        auto_mission: downloads the latest mission from Github
    """

    if CFG.sentry_dsn:
        from esst.utils.sentry import Sentry
        CTX.sentry = Sentry(CFG.sentry_dsn)
        CTX.sentry.register_context('App context', CTX)
        CTX.sentry.register_context('Config', CFG)

    CTX.loop = asyncio.get_event_loop()

    import esst.wan
    CTX.wan = CTX.loop.run_until_complete(esst.wan.wan_available())
    CTX.loop.create_task(esst.wan.monitor_connection())

    CTX.start_discord_loop = discord and CFG.start_discord_loop
    CTX.start_server_loop = server and CFG.start_server_loop
    CTX.start_dcs_loop = dcs and CFG.start_dcs_loop
    CTX.start_listener_loop = listener and CFG.start_listener_loop

    if not (start_dcs and CFG.dcs_can_start):
        CTX.dcs_blocker.append('config')

    CTX.dcs_setup_dedi_config = install_dedi_config
    CTX.dcs_install_hooks = install_hooks
    CTX.dcs_auto_mission = auto_mission

    CTX.loop = asyncio.get_event_loop()
    # CTX.loop.set_debug(True)
    CTX.discord_msg_queue = queue.Queue()

    import ctypes
    ctypes.windll.kernel32.SetConsoleTitleW(f'ESST v{__version__} - Use CTRL+C to exit')
    MAIN_LOGGER.debug(f'starting ESST {__version__}')

    fs_paths.init_fs(CFG)

    from esst.utils import clean_all_folder
    clean_all_folder()

    from esst import atis
    atis.init_module()

    import esst.discord_bot.discord_bot
    discord_loop = esst.discord_bot.discord_bot.App()

    import esst.dcs

    dcs_loop = esst.dcs.dcs.App()

    import esst.server.server
    server_loop = esst.server.server.App()

    from esst.listener import DCSListener
    listener_loop = DCSListener()

    futures = asyncio.gather(
        CTX.loop.create_task(discord_loop.run()),
        CTX.loop.create_task(dcs_loop.run()),
        CTX.loop.create_task(listener_loop.run()),
        CTX.loop.create_task(server_loop.run()),
        CTX.loop.create_task(watch_for_exceptions()),
    )

    def sigint_handler(*_):
        """
        Catches exit signal (triggered byu CTRL+C)

        Args:
            *_: frame

        """
        MAIN_LOGGER.info(
            'ESST has been interrupted by user request, shutting down')
        CTX.exit = True

    import signal
    signal.signal(signal.SIGINT, sigint_handler)
    CTX.loop.run_until_complete(futures)
    MAIN_LOGGER.debug('main loop is done, killing DCS')

    futures = asyncio.gather(
        CTX.loop.create_task(dcs_loop.kill_running_app()),
        CTX.loop.create_task(listener_loop.run_until_dcs_is_closed()),
    )

    CTX.loop.run_until_complete(futures)
    if CTX.restart and CFG.restart:
        MAIN_LOGGER.debug(f'restart command: {CFG.restart}')
        bat = Path('./restart.bat')
        bat.write_text(CFG.restart)
        os.startfile(str(bat.absolute()))
    MAIN_LOGGER.debug('all done !')


if __name__ == '__main__':
    main()  # pylint: disable=no-value-for-parameter



# coding=utf-8
"""
Meh
"""
import pprint
from time import sleep

import emiz.edit_miz
import emiz.weather

from esst import commands, core, dcs, utils

LOGGER = core.MAIN_LOGGER.getChild(__name__)


def _mission_index_to_mission_name(mission_index):
    LOGGER.debug(f'converting mission index to mission name: {mission_index}')
    for index, mission_name in dcs.missions_manager.list_available_missions():
        if index == mission_index:
            LOGGER.debug(f'mission found: {mission_name}')
            return dcs.missions_manager.MissionPath(mission_name)
    LOGGER.debug('no mission found')
    return None


# pylint: disable=too-many-statements,too-many-branches,too-many-return-statements,too-many-arguments
def _load(name, icao, metar, time, max_wind, min_wind, force):  # noqa: C901
    if name is None:
        mission = dcs.missions_manager.get_running_mission()
        if not mission:
            LOGGER.error('unable to retrieve current mission')
            return
    else:
        try:
            LOGGER.debug(f'trying to cast mission name into an int: {name}')
            mission_number = int(name)
        except ValueError:
            LOGGER.debug(f'loading mission name: {name}')
            mission = dcs.missions_manager.MissionPath(name)
            if not mission:
                LOGGER.debug(f'mission path not found: {mission.path}')
                LOGGER.error(f'mission file not found: {mission.name}')
                return
        else:
            LOGGER.debug(f'loading mission number: {mission_number}')
            mission = _mission_index_to_mission_name(mission_number)
            if not mission:
                LOGGER.error(f'invalid mission index: {mission_number}; use "!mission  show" to see available indices')
                return

    LOGGER.info(f'loading mission file: {mission.path}')
    if metar:
        metar = ' '.join(metar)
        LOGGER.info(f'analyzing METAR string: {metar}')
        error, metar = emiz.weather.custom_metar.CustomMetar.get_metar(metar)
        if error:
            LOGGER.error(error)
            return
    if icao:
        icao = icao.upper()
        LOGGER.info(f'obtaining METAR from: {icao}')
        error, metar_str = emiz.weather.noaa.retrieve_metar(icao)
        if error:
            LOGGER.error(error)
            return
        LOGGER.info(f'analyzing METAR string: {metar_str}')
        error, metar = emiz.weather.custom_metar.CustomMetar.get_metar(metar_str)
        if error:
            LOGGER.error(error)
            return

    if metar:
        info_metar = metar
        LOGGER.info(f'METAR: {metar.string()}')
    else:
        LOGGER.info('building METAR from mission file')
        # noinspection SpellCheckingInspection
        metar_str = emiz.weather.mizfile.get_metar_from_mission(str(mission.path), 'XXXX')
        error, info_metar = emiz.weather.custom_metar.CustomMetar.get_metar(metar_str)
        if error:
            LOGGER.error(error)
            return
        LOGGER.info(f'METAR: {info_metar.string()}')

    LOGGER.debug(f'editing "{mission.path}" to "{mission.auto.path}"')
    commands.DCS.block_start('loading mission')
    commands.DCS.kill(force=force)
    LOGGER.debug('waiting on DCS application to close')
    while core.Status.dcs_application != 'not running':
        sleep(1)
    LOGGER.debug('DCS has closed, carrying on')
    edit_str = []
    if time:
        edit_str.append('time')
    if metar:
        edit_str.append('weather')
    if edit_str:
        edit_str = ' and '.join(edit_str)
        LOGGER.info(
            f'loading {mission.name} with {edit_str} (this may take a few seconds)')
    else:
        LOGGER.info(f'loading {mission.name} as is (no edit)')
    try:
        miz_edit_options = dict(infile=str(mission.path), outfile=str(mission.auto.path), metar=metar, time=time,
                                min_wind=min_wind, max_wind=max_wind)
        LOGGER.debug(f'editing miz file with options:\n{pprint.pformat(miz_edit_options)}')
        error = emiz.edit_miz.edit_miz(**miz_edit_options)
        if error:
            if error == 'nothing to do!':
                LOGGER.debug(f'loading mission "as is": {mission.path}')
                mission.set_as_active(info_metar.code)
            else:
                LOGGER.error(error)
        else:
            LOGGER.debug(f'mission has been successfully edited, setting as active: {mission.auto.path}')
            mission.auto.set_as_active(info_metar.code)
    finally:
        commands.DCS.unblock_start('loading mission')


@utils.arg(protected=True)
def delete(name: 'name or index of the mission to load'):
    """
    Removes a mission file from the server (protected)
    """
    try:
        mission_number = int(name)
    except ValueError:
        mission = dcs.missions_manager.MissionPath(name)
        if not mission:
            LOGGER.error(f'mission file does not exist: {mission.path}')
            return
    else:
        mission = _mission_index_to_mission_name(mission_number)
        if not mission:
            LOGGER.error(
                f'invalid mission index: {mission_number}; use "!mission  show" to see available indices')
            return

    dcs.missions_manager.delete(mission)


# noinspection SpellCheckingInspection
@utils.arg('-m', '--metar', nargs='+', metavar='METAR')
@utils.arg(protected=True)
# pylint: disable=too-many-arguments
def load(
        name: 'name or index of the mission to load (if not provided, will re-use the current mission)' = None,
        icao: 'update the weather from ICAO' = None,
        metar: 'update the weather from METAR string\nWARNING: METAR string may NOT contain dashes ("-")' = None,
        time: 'set the mission time (syntax: YYYYMMDDHHMMSS)\nEx: 2017/08/22 at 12:30:00 -> 20170822123000' = None,
        max_wind: 'maximum speed of the wind in MPS' = 40,
        min_wind: 'minimum speed of the wind in MPS' = 0,
        force: 'force server restart even with connected players' = False,

):
    """
    Load a mission, allowing to set the weather or the time (protected).

    Missions can be loaded by typing their full name, or using the index number given by the "!mission show" command.
    """
    if not (force or commands.DCS.check_for_connected_players()):
        return
    if not any((name, icao, metar, time)):
        commands.DISCORD.say('Type "!mission load --help" to see available options')
        return
    core.CTX.loop.run_in_executor(None, _load, name, icao,
                                  metar, time, max_wind, min_wind, force)


def show():
    """
    Show list of missions available on the server
    """
    available_mission = '\n\t'.join(
        f'{n}. {m}' for n, m in dcs.missions_manager.list_available_missions())
    # available_mission = '\n\t'.join(available_mission)
    commands.DISCORD.say(
        'Available missions:\n'
        f'\t{available_mission}\n'
    )


def weather():
    """
    Displays the weather for the currently running mission
    """
    if core.Status.metar and core.Status.metar != 'unknown':
        error, metar = emiz.weather.custom_metar.CustomMetar.get_metar(core.Status.metar)
        if error:
            LOGGER.error(error)
            return
        else:
            commands.DISCORD.say(f'{metar.string()}')


def download():
    """
    Sends the currently running mission on Discord
    """
    mission = dcs.missions_manager.get_running_mission()
    if mission:
        commands.DISCORD.send_file(str(mission.path))


NAMESPACE = '!mission'
TITLE = 'Manage missions'



# coding=utf-8
"""
Rotates dcs.log files
"""

import datetime
from pathlib import Path

from esst.core import CFG, FS, MAIN_LOGGER
from esst.utils.remove_old_files import parse_age_string, remove_file_if_older_than

LOGGER = MAIN_LOGGER.getChild(__name__)


def _log_dir() -> Path:
    return FS.dcs_logs_dir


def _save_old_log(old_log: Path):
    stat = old_log.stat()
    creation_time = datetime.datetime.fromtimestamp(stat.st_mtime)
    old_log.rename(
        Path(old_log.parent, f'{creation_time.strftime("%Y%m%d%H%M%S")}.dcs.log'))


def rotate_dcs_log():
    """
    Rotates DCS logs
    """
    LOGGER.info('rotating DCS logs')
    log_dir = _log_dir()
    LOGGER.debug(f'using logs directory: {log_dir}')
    if not log_dir.exists():
        LOGGER.error('log directory does no exist')
        return
    old_log = Path(log_dir.joinpath('dcs.log.old'))
    if old_log.exists():
        LOGGER.debug('saving old log')
        _save_old_log(old_log)
    else:
        LOGGER.info('no old log found')


def clean_old_logs():
    """
    Removes old logs
    """
    if CFG.dcs_delete_logs_older_than:
        age = parse_age_string(CFG.dcs_delete_logs_older_than)
        if not age:
            LOGGER.error(
                f'invalid value for "dcs_keep_logs_for": {CFG.dcs_delete_logs_older_than}')
            return
        LOGGER.info('removing old DCS logs')
        log_dir = _log_dir()
        LOGGER.debug(f'using logs directory: {log_dir}')
        if not log_dir.exists():
            LOGGER.error('log directory does no exist')
            return
        for file in log_dir.glob('*.dcs.log'):
            remove_file_if_older_than(file, age)
    else:
        LOGGER.info('not removing old logs; no age limit given in config')



# coding=utf-8
"""
Provides random phonetic letter for ATIS identifier
"""

import string

from elib.custom_random import random_string

# noinspection SpellCheckingInspection
PHONETIC = {'A': 'Alfa', 'B': 'Bravo', 'C': 'Charlie', 'D': 'Del-tah', 'E': 'Echo', 'F': 'Foxtrot', 'G': 'Golf',
            "H": "Hotel", 'I': 'India', 'J': 'Juliet', 'K': 'Keelo', 'L': 'Leema', 'M': 'Mike', 'N': 'November',
            'O': 'Oscar', 'P': 'Papa', 'Q': 'Quebec', 'R': 'Romeo', 'S': 'Sierra', 'T': 'Tango', 'U': 'Uniform',
            'V': 'Victor', 'W': 'Whiskey', 'X': 'ex-ray', 'Y': 'Yankee', 'Z': 'Zulu'}


def get_random_identifier():
    """

    Returns: random identifier for ATIS

    """
    letter = random_string(1, string.ascii_uppercase)
    return PHONETIC[letter], letter



# coding=utf-8
"""
Set up Config object
"""

import everett

from ._new_config import ESSTConfig
from ._validate_config import validate_config


def setup_config():
    """
    Set up the config object

    Returns: instance of ESSTConfig

    """
    try:
        config = ESSTConfig()
        validate_config(config)
    except everett.InvalidValueError as exception:
        key = exception.key
        if exception.namespace:
            key = f'{exception.namespace}_{key}'
        print(f'Invalid config value: {key}')
        exit(1)
    except everett.ConfigurationMissingError as exception:
        key = exception.key
        if exception.namespace:
            key = f'{exception.namespace}_{key}'
        print(f'Missing configuration value: {key}')
        exit(1)
    else:
        return config



# coding=utf-8
"""
Declares all airfields in Caucasus
"""
from ._univers_radio import Airfield, URCoord, URFrequency

# noinspection SpellCheckingInspection
UG5X = Airfield('UG5X', 'Kobuleti',
                URCoord('41.929501', '41.871101', '59'),
                URFrequency('133.400;A'),
                '07,25')
# noinspection SpellCheckingInspection
UG23 = Airfield('UG23', 'Gudauta',
                URCoord('43.101501', '40.581299', '68'),
                URFrequency('130.400;A'),
                '15,33')
# noinspection SpellCheckingInspection
UG24 = Airfield('UG24', 'Tbilisi - Soganlug',
                URCoord('41.649200', '44.933701', '1472'),
                URFrequency('139.400;A'),
                '13,31')
# noinspection SpellCheckingInspection
UG27 = Airfield('UG27', 'Vaziani',
                URCoord('41.628300', '45.030201', '1492'),
                URFrequency('140.400;A'),
                '13,31')
# noinspection SpellCheckingInspection
UGKO = Airfield('UGKO', 'Kutaisi - Kopitnari',
                URCoord('42.178699', '42.483101', '147'),
                URFrequency('134.400;A'),
                '07,25')
# noinspection SpellCheckingInspection
UGKS = Airfield('UGKS', 'Senaki - Kolkhi',
                URCoord('42.240101', '42.052101', '42'),
                URFrequency('132.400;A'),
                '09,17')
# noinspection SpellCheckingInspection
UGSB = Airfield('UGSB', 'Batumi',
                URCoord('41.610802', '41.598202', '32'),
                URFrequency('131.400;A'),
                '12,30')
# noinspection SpellCheckingInspection
UGSS = Airfield('UGSS', 'Sukhumi - Babushara',
                URCoord('42.867100', '41.120602', '29'),
                URFrequency('129.400;A'),
                '11,29')
# noinspection SpellCheckingInspection
UGTB = Airfield('UGTB', 'Tbilisi - Lochini',
                URCoord('41.673199', '44.953499', '1538'),
                URFrequency('138.400;A'),
                '13R,31L')
# noinspection SpellCheckingInspection
URKA = Airfield('URKA', 'Anapa - Vityazevo',
                URCoord('45.002899', '37.340801', '147'),
                URFrequency('121.400;A'),
                '04,22')
# noinspection SpellCheckingInspection
URKG = Airfield('URKG', 'Geledzhik',
                URCoord('44.592300', '38.023300', '82'),
                URFrequency('126.400;A'),
                '04,22')
# noinspection SpellCheckingInspection
URKH = Airfield('URKH', 'Maykop - Khanskaya',
                URCoord('44.681400', '40.032501', '590'),
                URFrequency('125.400;A'),
                '04,22')
# noinspection SpellCheckingInspection
URKK = Airfield('URKK', 'Krasnodar - Pashkovsky',
                URCoord('45.035999', '39.146702', '111'),
                URFrequency('128.400;A'),
                '04L,22R,04R,22L')
# noinspection SpellCheckingInspection
URKL = Airfield('URKL', 'Krasnodar - Center',
                URCoord('45.080700', '38.961899', '98'),
                URFrequency('122.400;A'),
                '08,26')
# noinspection SpellCheckingInspection
URKN = Airfield('URKN', 'Novorossiysk',
                URCoord('44.678398', '37.782501', '131'),
                URFrequency('123.400;A'),
                '04,22')
# noinspection SpellCheckingInspection
URKW = Airfield('URKW', 'Krymsk',
                URCoord('44.958500', '37.996601', '65'),
                URFrequency('124.400;A'),
                '03,21')
# noinspection SpellCheckingInspection
URMM = Airfield('URMM', 'Mineralnye Vody',
                URCoord('44.219501', '43.085899', '1049'),
                URFrequency('135.400;A'),
                '11,29')
# noinspection SpellCheckingInspection
URMN = Airfield('URMN', 'Nalchik',
                URCoord('43.511200', '43.635201', '1410'),
                URFrequency('136.400;A'),
                '05,23')
# noinspection SpellCheckingInspection
URMO = Airfield('URMO', 'Beslan',
                URCoord('43.203899', '44.608398', '1908'),
                URFrequency('141.400;A'),
                '09,27')
# noinspection SpellCheckingInspection
URSS = Airfield('URSS', 'Sochi - Adler',
                URCoord('43.443501', '39.941502', '98'),
                URFrequency('127.400;A'),
                '02,20,06,24')
# noinspection SpellCheckingInspection
XRMF = Airfield('XRMF', 'Mozdok',
                URCoord('43.786201', '44.607201', '508'),
                URFrequency('137.400;A'),
                '08,26')

ALL_AIRFIELDS = [UG23, UG24, UG27, UG5X, UGKO, UGKS, UGSB, UGSS, UGTB, URKG,
                 URKH, URKK, URKL, URKN, URKW, URMM, URMN, URMO, URSS, XRMF]



# coding=utf-8
"""
Tests from esst.dcs.autoexec_cfg
"""

import string
from pathlib import Path

import pytest
from hypothesis import given
from hypothesis import strategies as st

from esst import core, dcs


@pytest.fixture(autouse=True)
def _setup():
    core.FS.saved_games_path = '.'
    core.FS.dcs_autoexec_file = Path('./DCS/Config/autoexec.cfg')


def test_injection():
    Path('./DCS/Config').mkdir(parents=True)
    autoexec_file = Path('./DCS/Config/autoexec.cfg')
    assert not autoexec_file.exists()
    dcs.autoexec_cfg.inject_silent_crash_report()
    assert autoexec_file.exists()
    assert autoexec_file.read_text('utf8').endswith(dcs.autoexec_cfg._SILENT_CRASH_REPORT)


def test_no_dcs_saved_games_path():
    core.FS.saved_games_path = None
    core.FS.dcs_autoexec_file = './autoexec.cfg'
    with pytest.raises(RuntimeError) as exc_info:
        dcs.autoexec_cfg.inject_silent_crash_report()

    assert 'path uninitialized: saved games' in str(exc_info)


def test_no_config_path():
    Path('./DCS').mkdir(parents=True)
    with pytest.raises(FileNotFoundError) as exc_info:
        dcs.autoexec_cfg.inject_silent_crash_report()

    assert 'Config' in str(exc_info)


@given(text=st.text(alphabet=string.printable, min_size=0, max_size=100))
def test_existing_file(text):
    Path('./DCS/Config').mkdir(parents=True, exist_ok=True)

    autoexec_file = Path('./DCS/Config/autoexec.cfg')
    autoexec_file.write_text(text, encoding='utf8')
    assert autoexec_file.exists()
    assert dcs.autoexec_cfg.inject_silent_crash_report()
    assert autoexec_file.exists()
    content = autoexec_file.read_text('utf8')
    assert content.endswith(dcs.autoexec_cfg._SILENT_CRASH_REPORT)
    assert dcs.autoexec_cfg.inject_silent_crash_report()
    # Make sure the content does not change
    assert autoexec_file.read_text('utf8') == content



# coding=utf-8
"""
Manages DCS commands
"""
import threading
import time
from queue import Queue

from esst.core import CTX, MAIN_LOGGER, Status

LOGGER = MAIN_LOGGER.getChild(__name__)
CANCEL_QUEUED_KILL = Queue()

_DCS_LOCK = threading.Lock()


class DCS:
    """
    Manages commands for the DCS application
    """

    @staticmethod
    def restart(force: bool = False):
        """
        Sets the context to restart the DCS application

        Returns: None if restart is OK, err as a str otherwise

        """
        CANCEL_QUEUED_KILL.put(1)
        if DCS.there_are_connected_players() and not force:
            LOGGER.error('there are connected players; cannot restart the server now '
                         ' (use "--force" to kill anyway)')
            return
        LOGGER.debug('setting context for DCS restart')
        CTX.dcs_do_restart = True

    @staticmethod
    def kill(force: bool = False, queue: bool = False):
        """Kills DCS application"""
        CANCEL_QUEUED_KILL.put(1)
        if DCS.there_are_connected_players():
            if not force:
                if queue:
                    DCS.queue_kill()
                else:
                    LOGGER.error('there are connected players; cannot kill the server now'
                                 ' (use "--force" to kill anyway)')
                return
            else:
                LOGGER.warning('forcing kill with connected players')
        LOGGER.debug('setting context for DCS kill')
        CTX.dcs_do_kill = True

    @staticmethod
    def queue_kill():
        """Kill DCS application when all players left"""
        def _queue_kill(queue: Queue):
            while DCS.there_are_connected_players():
                if not queue.empty():
                    queue.get_nowait()
                    LOGGER.debug('queued DCS kill has been cancelled')
                    return
                time.sleep(5)
            LOGGER.info('executing planned DCS restart')
            DCS.kill()

        LOGGER.warning('queuing DCS kill for when all players have left')
        CTX.loop.run_in_executor(None, _queue_kill, CANCEL_QUEUED_KILL)

    @staticmethod
    def show_cpu_usage_once():
        """Show CPU usage once"""
        LOGGER.debug('show cpu usage once')
        CTX.dcs_show_cpu_usage_once = True

    @staticmethod
    def show_cpu_usage_once_done():
        """Show CPU usage once"""
        LOGGER.debug('show cpu usage once: done')
        CTX.dcs_show_cpu_usage_once = False

    @staticmethod
    def show_cpu_usage_start():
        """Starts showing CPU usage continuously"""
        LOGGER.debug('show cpu usage: start')
        CTX.dcs_show_cpu_usage = True

    @staticmethod
    def show_cpu_usage_stop():
        """Stops showing CPU usage continuously"""
        LOGGER.debug('show cpu usage: stop')
        CTX.dcs_show_cpu_usage = False

    @staticmethod
    def block_start(reason: str):
        """
        Prevents DCS from starting

        Args:
            reason: reason for preventing DCS to start
        """
        with _DCS_LOCK:
            CTX.dcs_blocker.append(reason)

    @staticmethod
    def unblock_start(reason: str):
        """
        Allow DCS to start

        Args:
            reason: reason that was preventing DCS to start
        """
        with _DCS_LOCK:
            if reason in CTX.dcs_blocker:
                CTX.dcs_blocker.remove(reason)

    @staticmethod
    def dcs_cannot_start() -> list:
        """Returns the list of reasons preventing DCS to start"""
        return CTX.dcs_blocker

    @staticmethod
    def there_are_connected_players() -> bool:
        """

        Returns: bool indicating if there are connected players
        """
        connected_players = bool(Status.players)
        if connected_players:
            LOGGER.debug(f'there are {len(Status.players)} connected player(s)')
        else:
            LOGGER.debug('there is no connected players')
        return connected_players

    @staticmethod
    def check_for_connected_players() -> bool:
        """
        Returns: False if there are connected players
        """
        if DCS.there_are_connected_players():
            LOGGER.warning('there are connected players; cannot kill the server now')
            return False

        return True



# coding=utf-8
"""
Manages config params for dcs server
"""
from elib.config import ConfigProp

NAMESPACE = 'DCS_SERVER'


class DCSServerConfig:
    """
    Manages config params for dcs server
    """
    dcs_server_name = ConfigProp(str, default='', namespace=NAMESPACE)
    dcs_server_max_players = ConfigProp(int, default='30', namespace=NAMESPACE)
    dcs_server_startup_time = ConfigProp(int, default='120', namespace=NAMESPACE)
    dcs_server_event_role = ConfigProp(str, default='true', namespace=NAMESPACE)
    dcs_server_require_pure_clients = ConfigProp(str, default='false', namespace=NAMESPACE)
    dcs_server_allow_ownship_export = ConfigProp(str, default='true', namespace=NAMESPACE)
    dcs_server_allow_object_export = ConfigProp(str, default='true', namespace=NAMESPACE)
    dcs_server_password = ConfigProp(str, default='', namespace=NAMESPACE)



# coding=utf-8
"""
Manages ATIS objects
"""

from attr import attrib, attrs


@attrs
class ATISForAirfield:
    """Simple data class for airfield ATIS information"""
    icao = attrib()
    active_runway = attrib()
    info_id = attrib()
    info_letter = attrib()



# coding=utf-8
"""
Dummy test to make sure everything is importable
"""

import glob

import pytest


@pytest.mark.nocleandir
@pytest.mark.parametrize('module_', glob.glob('./esst/**/*.py', recursive=True))
def test_imports(module_):
    module_ = module_[2:-3].replace('\\', '.')
    __import__(module_)


@pytest.mark.nocleandir
@pytest.mark.parametrize('module_', list(glob.glob('./esst/**/*.py', recursive=True)))
def test_imports_tests(module_):
    module_ = module_[2:-3].replace('\\', '.')
    __import__(module_)



# coding=utf-8
"""
Installs the necessary files to run in dedicated mode
"""
from pathlib import Path

import jinja2

from esst.core import CFG, CTX, FS, MAIN_LOGGER
from esst.utils import create_versioned_backup, read_template

LOGGER = MAIN_LOGGER.getChild(__name__)

DEDI_CFG = r"""dedicated =
{
    ["enabled"] = true,
}
"""


def _get_me_auth_path() -> Path:
    me_auth_path = Path(CFG.dcs_path, 'MissionEditor/modules/me_authorization.lua')
    if not me_auth_path.exists():
        raise FileNotFoundError(str(me_auth_path))
    return me_auth_path


def _write_dedi_config():
    dedi_cfg_path = Path(FS.variant_saved_games_path, 'Config/dedicated.lua')
    if not dedi_cfg_path.exists():
        LOGGER.info(f'writing {dedi_cfg_path}')
        dedi_cfg_path.write_text(DEDI_CFG)
    else:
        LOGGER.debug(f'file already exists: {dedi_cfg_path}')


def _write_auth_file():
    content = read_template('me_authorization.lua')
    LOGGER.debug('writing me_authorization.lua')
    _get_me_auth_path().write_text(jinja2.Template(content).render(server_name=CFG.discord_bot_name))


def setup_config_for_dedicated_run():
    """
    Setup the server to automatically starts in multiplayer mode when DCS starts
    """
    if CTX.dcs_setup_dedi_config:
        LOGGER.debug('setting up dedicated config')
        create_versioned_backup(_get_me_auth_path())
        _write_auth_file()
        _write_dedi_config()
        LOGGER.debug('setting up dedicated config: all done!')
    else:
        LOGGER.debug('skipping installation of dedicated config')



# coding=utf-8
"""
Manages config params for auto mission
"""

from elib.config import ConfigProp

from esst import __version__

NAMESPACE = 'DISCORD'

DEFAULT_MOTD = f"""
Hello!

This is ESST v{__version__}
Type "!help" for a list of available commands
"""


class DiscordConfig:
    """
    Manages config params for auto mission
    """
    discord_bot_name = ConfigProp(str, default='', namespace=NAMESPACE)
    discord_channel = ConfigProp(str, default='', namespace=NAMESPACE)
    discord_token = ConfigProp(str, default='', namespace=NAMESPACE)
    discord_admin_role = ConfigProp(list, default='', namespace=NAMESPACE)
    discord_motd = ConfigProp(str, namespace=NAMESPACE, default=DEFAULT_MOTD)
